# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import pytest
from craft_store.models import release_request_model

from .conftest import needs_charmhub_credentials

pytestmark = pytest.mark.timeout(10)  # Timeout if any test takes over 10 sec.


@needs_charmhub_credentials()
def test_charm_release(charm_client, charmhub_charm_name):
    model = release_request_model.ReleaseRequestModel(
        channel="edge", revision=1, resources=[]
    )

    charm_client.release(
        name=charmhub_charm_name,
        release_request=[model],
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Model for an error from the craft store."""
from craft_store.models._base_model import MarshableModel


class ErrorModel(MarshableModel):
    """An error."""

    code: str
    message: str



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from unittest.mock import Mock, call, patch

import pytest
from craft_store import creds, endpoints, errors
from craft_store.ubuntu_one_store_client import UbuntuOneStoreClient
from pymacaroons import Caveat, Macaroon


def _fake_response(status_code, reason=None, headers=None, json=None):
    response = Mock(spec="requests.Response")
    response.status_code = status_code
    response.ok = status_code == 200
    response.headers = headers
    response.reason = reason
    if json is not None:
        response.json = Mock(return_value=json)
    return response


@pytest.fixture()
def root_macaroon():
    return Macaroon(
        location="fake-server.com",
        signature="d9533461d7835e4851c7e3b639144406cf768597dea6e133232fbd2385a5c050",
        caveats=[
            Caveat(
                caveat_id="1234567890",
                location="fake-sso.com",
                verification_key_id="1234567890",
            )
        ],
    ).serialize()


@pytest.fixture()
def discharged_macaroon():
    return Macaroon(
        location="fake-server.com",
        signature="d9533461d7835e4851c7e3b639122406cf768597dea6e133232fbd2385a5c050",
    ).serialize()


@pytest.fixture()
def u1_macaroon_value(root_macaroon, discharged_macaroon):
    """The basic "payload" for the u1-macaroon auth type"""
    return {"r": root_macaroon, "d": discharged_macaroon}


@pytest.fixture()
def old_credentials(u1_macaroon_value):
    """u1-macaroon credentials encoded in the *old* ("type-less") scheme."""
    return json.dumps(u1_macaroon_value)


@pytest.fixture()
def new_credentials(u1_macaroon_value):
    """u1-macaroon credentials encoded in the *new* ("typed") scheme."""
    return creds.marshal_u1_credentials(creds.UbuntuOneMacaroons(**u1_macaroon_value))


@pytest.fixture()
def authorization():
    return (
        "Macaroon "
        "root=MDAxZGxvY2F0aW9uIGZha2Utc2VydmVyLmNvbQowMDEwaWRlbnRpZmllciAKMDAxM2NpZCAxMjM0NTY3ODkwCjAwMTN2aWQgMTIzNDU2Nzg5MAowMDE0Y2wgZmFrZS1zc28uY29tCjAwMmZzaWduYXR1cmUg2VM0YdeDXkhRx-O2ORREBs92hZfepuEzIy-9I4WlwFAK, "
        "discharge=MDAxZGxvY2F0aW9uIGZha2Utc2VydmVyLmNvbQowMDEwaWRlbnRpZmllciAKMDAyZnNpZ25hdHVyZSB6hf06Su8kgum0keaUXy6VxGUHlN9bFL2A0EKNptFZMwo"
    )


@pytest.fixture()
def http_client_request_mock(root_macaroon, discharged_macaroon):
    def request(*args, **kwargs):
        if args[1] == "POST" and "tokens/discharge" in args[2]:
            email = kwargs["json"]["email"]
            otp = kwargs["json"].get("otp")
            if email == "otp@foo.bar" and otp is None:
                response = _fake_response(
                    401,
                    json={
                        "error_list": [
                            {"code": "twofactor-required", "message": "otp required"}
                        ]
                    },
                )
            else:
                response = _fake_response(
                    200,
                    json={"discharge_macaroon": discharged_macaroon},
                )
        elif args[1] == "POST" and "/tokens/refresh" in args[2]:
            response = _fake_response(
                200,
                json={"discharge_macaroon": discharged_macaroon},
            )
        elif args[1] == "POST" and "/dev/api/acl/" in args[2]:
            response = _fake_response(200, json={"macaroon": root_macaroon})
        elif args[1] == "GET" and "whoami" in args[2]:
            response = _fake_response(
                200,
                json={"name": "Fake Person", "username": "fakeuser", "id": "fake-id"},
            )
        else:
            response = _fake_response(200)

        return response

    patched_http_client = patch(
        "craft_store.base_client.HTTPClient.request",
        autospec=True,
        side_effect=request,
    )
    mocked_http_client = patched_http_client.start()
    yield mocked_http_client
    patched_http_client.stop()


@pytest.fixture()
def auth_mock(old_credentials, new_credentials, new_auth):
    patched_auth = patch("craft_store.base_client.Auth", autospec=True)
    mocked_auth = patched_auth.start()

    credentials = new_credentials if new_auth else old_credentials

    mocked_auth.return_value.get_credentials.return_value = credentials
    # Note that the call to encode credentials always encode the new format.
    mocked_auth.return_value.encode_credentials.return_value = new_credentials
    yield mocked_auth
    patched_auth.stop()


@pytest.mark.parametrize("environment_auth", [None, "APPLICATION_CREDENTIALS"])
def test_store_client_login(
    http_client_request_mock,
    new_credentials,
    auth_mock,
    environment_auth,
    expires,
):
    store_client = UbuntuOneStoreClient(
        base_url="https://fake-server.com",
        storage_base_url="https://fake-storage.com",
        auth_url="https://fake-sso.com",
        endpoints=endpoints.U1_SNAP_STORE,
        application_name="fakecraft",
        user_agent="FakeCraft Unix X11",
        environment_auth=environment_auth,
    )

    assert (
        store_client.login(
            permissions=["perm-1", "perm-2"],
            description="fakecraft@foo",
            ttl=60,
            email="foo@bar.com",
            password="password",  # noqa: S106
        )
        == new_credentials
    )

    assert http_client_request_mock.mock_calls == [
        call(
            store_client.http_client,
            "POST",
            "https://fake-server.com/dev/api/acl/",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            json={
                "permissions": ["perm-1", "perm-2"],
                "description": "fakecraft@foo",
                "expires": expires(60),
            },
        ),
        call(
            store_client.http_client,
            "POST",
            "https://fake-sso.com/api/v2/tokens/discharge",
            json={
                "email": "foo@bar.com",
                "password": "password",
                "caveat_id": "1234567890",
            },
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        ),
    ]

    assert auth_mock.mock_calls == [
        call(
            "fakecraft",
            "fake-server.com",
            environment_auth=environment_auth,
            ephemeral=False,
        ),
        call().ensure_no_credentials(),
        call().set_credentials(new_credentials),
        call().encode_credentials(new_credentials),
    ]


def test_store_client_login_otp(
    http_client_request_mock,
    new_credentials,
    auth_mock,
    expires,
):
    store_client = UbuntuOneStoreClient(
        base_url="https://fake-server.com",
        storage_base_url="https://fake-storage.com",
        auth_url="https://fake-sso.com",
        endpoints=endpoints.U1_SNAP_STORE,
        application_name="fakecraft",
        user_agent="FakeCraft Unix X11",
    )

    with pytest.raises(errors.StoreServerError) as server_error:
        store_client.login(
            permissions=["perm-1", "perm-2"],
            description="fakecraft@foo",
            ttl=60,
            email="otp@foo.bar",
            password="password",  # noqa: S106
        )
    assert "twofactor-required" in server_error.value.error_list

    assert (
        store_client.login(
            permissions=["perm-1", "perm-2"],
            description="fakecraft@foo",
            ttl=60,
            email="otp@foo.bar",
            password="password",  # noqa: S106
            otp="123456",
        )
        == new_credentials
    )

    assert http_client_request_mock.mock_calls == [
        call(
            store_client.http_client,
            "POST",
            "https://fake-server.com/dev/api/acl/",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            json={
                "permissions": ["perm-1", "perm-2"],
                "description": "fakecraft@foo",
                "expires": expires(60),
            },
        ),
        call(
            store_client.http_client,
            "POST",
            "https://fake-sso.com/api/v2/tokens/discharge",
            json={
                "email": "otp@foo.bar",
                "password": "password",
                "caveat_id": "1234567890",
            },
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        ),
        call(
            store_client.http_client,
            "POST",
            "https://fake-server.com/dev/api/acl/",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            json={
                "permissions": ["perm-1", "perm-2"],
                "description": "fakecraft@foo",
                "expires": expires(60),
            },
        ),
        call(
            store_client.http_client,
            "POST",
            "https://fake-sso.com/api/v2/tokens/discharge",
            json={
                "email": "otp@foo.bar",
                "password": "password",
                "otp": "123456",
                "caveat_id": "1234567890",
            },
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        ),
    ]

    assert auth_mock.mock_calls == [
        call("fakecraft", "fake-server.com", environment_auth=None, ephemeral=False),
        # First call without otp.
        call().ensure_no_credentials(),
        # Second call with otp.
        call().ensure_no_credentials(),
        call().set_credentials(new_credentials),
        call().encode_credentials(new_credentials),
    ]


def test_store_client_login_with_packages_and_channels(
    http_client_request_mock, new_credentials, auth_mock, expires
):
    store_client = UbuntuOneStoreClient(
        base_url="https://fake-server.com",
        storage_base_url="https://fake-storage.com",
        auth_url="https://fake-sso.com",
        endpoints=endpoints.U1_SNAP_STORE,
        application_name="fakecraft",
        user_agent="FakeCraft Unix X11",
    )

    assert (
        store_client.login(
            permissions=["perm-1", "perm-2"],
            description="fakecraft@foo",
            ttl=60,
            channels=["edge"],
            packages=[
                endpoints.Package("my-snap", "snap"),
                endpoints.Package("my-other-snap", "snap"),
            ],
            email="foo@bar.com",
            password="password",  # noqa: S106
        )
        == new_credentials
    )

    assert http_client_request_mock.mock_calls == [
        call(
            store_client.http_client,
            "POST",
            "https://fake-server.com/dev/api/acl/",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            json={
                "permissions": ["perm-1", "perm-2"],
                "description": "fakecraft@foo",
                "expires": expires(60),
                "packages": [
                    {
                        "name": "my-snap",
                        "series": "16",
                    },
                    {
                        "name": "my-other-snap",
                        "series": "16",
                    },
                ],
                "channels": ["edge"],
            },
        ),
        call(
            store_client.http_client,
            "POST",
            "https://fake-sso.com/api/v2/tokens/discharge",
            json={
                "email": "foo@bar.com",
                "password": "password",
                "caveat_id": "1234567890",
            },
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        ),
    ]

    assert auth_mock.mock_calls == [
        call("fakecraft", "fake-server.com", environment_auth=None, ephemeral=False),
        call().ensure_no_credentials(),
        call().set_credentials(new_credentials),
        call().encode_credentials(new_credentials),
    ]


def test_store_client_logout(auth_mock):
    store_client = UbuntuOneStoreClient(
        base_url="https://fake-server.com",
        storage_base_url="https://fake-storage.com",
        auth_url="https://fake-sso.com",
        endpoints=endpoints.U1_SNAP_STORE,
        application_name="fakecraft",
        user_agent="FakeCraft Unix X11",
    )

    store_client.logout()

    assert auth_mock.mock_calls == [
        call("fakecraft", "fake-server.com", environment_auth=None, ephemeral=False),
        call().del_credentials(),
    ]


def test_store_client_request(http_client_request_mock, authorization, auth_mock):
    store_client = UbuntuOneStoreClient(
        base_url="https://fake-server.com",
        storage_base_url="https://fake-storage.com",
        auth_url="https://fake-sso.com",
        endpoints=endpoints.U1_SNAP_STORE,
        application_name="fakecraft",
        user_agent="FakeCraft Unix X11",
    )

    store_client.request("GET", "https://fake-server.com/fakepath")

    assert http_client_request_mock.mock_calls == [
        call(
            store_client.http_client,
            "GET",
            "https://fake-server.com/fakepath",
            params=None,
            headers={"Authorization": authorization},
        )
    ]

    assert auth_mock.mock_calls == [
        call("fakecraft", "fake-server.com", environment_auth=None, ephemeral=False),
        call().get_credentials(),
    ]


def test_store_client_request_refresh(
    http_client_request_mock,
    new_credentials,
    authorization,
    discharged_macaroon,
    auth_mock,
):
    http_client_request_mock.side_effect = [
        errors.StoreServerError(
            _fake_response(
                401,
                json={
                    "error_list": [
                        {
                            "code": "macaroon-needs-refresh",
                            "message": "Expired macaroon (age: 1234567 seconds)",
                        }
                    ]
                },
            )
        ),
        _fake_response(200, json={"discharge_macaroon": discharged_macaroon}),
        _fake_response(200),
    ]
    store_client = UbuntuOneStoreClient(
        base_url="https://fake-server.com",
        storage_base_url="https://fake-storage.com",
        auth_url="https://fake-sso.com",
        endpoints=endpoints.U1_SNAP_STORE,
        application_name="fakecraft",
        user_agent="FakeCraft Unix X11",
    )

    store_client.request("GET", "https://fake-server.com/refreshpath")

    assert http_client_request_mock.mock_calls == [
        call(
            store_client.http_client,
            "GET",
            "https://fake-server.com/refreshpath",
            params=None,
            headers={"Authorization": authorization},
        ),
        call(
            store_client.http_client,
            "POST",
            "https://fake-sso.com/api/v2/tokens/refresh",
            json={"discharge_macaroon": discharged_macaroon},
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        ),
        call(
            store_client.http_client,
            "GET",
            "https://fake-server.com/refreshpath",
            params=None,
            headers={"Authorization": authorization},
        ),
    ]

    assert auth_mock.mock_calls == [
        call("fakecraft", "fake-server.com", environment_auth=None, ephemeral=False),
        call().get_credentials(),
        call().get_credentials(),
        call().set_credentials(new_credentials, force=True),
        call().get_credentials(),
    ]


def test_store_client_whoami(http_client_request_mock, authorization, auth_mock):
    store_client = UbuntuOneStoreClient(
        base_url="https://fake-server.com",
        storage_base_url="https://fake-storage.com",
        auth_url="https://fake-sso.com",
        endpoints=endpoints.U1_SNAP_STORE,
        application_name="fakecraft",
        user_agent="FakeCraft Unix X11",
    )

    assert store_client.whoami() == {
        "name": "Fake Person",
        "username": "fakeuser",
        "id": "fake-id",
    }

    assert http_client_request_mock.mock_calls == [
        call(
            store_client.http_client,
            "GET",
            "https://fake-server.com/api/v2/tokens/whoami",
            params=None,
            headers={
                "Authorization": authorization,
            },
        )
    ]

    assert auth_mock.mock_calls == [
        call("fakecraft", "fake-server.com", environment_auth=None, ephemeral=False),
        call().get_credentials(),
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Revisions response models for the Store."""
import datetime
from typing import Any

from craft_store.models._base_model import MarshableModel
from craft_store.models._charm_model import CharmBaseModel
from craft_store.models._snap_models import Confinement, Grade, Type
from craft_store.models.error_model import ErrorModel


class RevisionsRequestModel(MarshableModel):
    """Resource model for a ReleaseRequestModel.

    :param upload_id: the upload-id returned from the storage endpoint.
    """

    upload_id: str


class RevisionsResponseModel(MarshableModel):
    """Model for a revisions response.

    :param status-url: a URL to monitor the state of the posted revision.
    """

    status_url: str


class RevisionModel(MarshableModel):
    """Base model for all revision types."""

    created_at: datetime.datetime
    revision: int
    sha3_384: str
    status: str

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> "RevisionModel":
        """Unmarshal a revision model."""
        if "bases" in data:
            return CharmRevisionModel.model_validate(data)
        if "apps" in data:
            return SnapRevisionModel.model_validate(data)
        if "commit-id" in data:
            return GitRevisionModel.model_validate(data)
        return RevisionModel.model_validate(data)


class GitRevisionModel(RevisionModel):
    """A model for a repository commit based revision."""

    commit_id: str
    created_by: str


class CharmRevisionModel(RevisionModel):
    """A revision model for charm revisions."""

    bases: list[CharmBaseModel]
    errors: list[ErrorModel] | None = None
    size: int
    version: str


class SnapRevisionModel(RevisionModel):
    """A model for a snap revision."""

    apps: list[str] | None = None
    architectures: list[str]
    base: str | None = None
    build_url: str | None = None
    confinement: Confinement
    created_by: str
    grade: Grade
    size: int
    type: Type
    version: str



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for list_releases."""
import datetime
from typing import cast

from craft_store.models import revisions_model
from craft_store.models._charm_model import CharmBaseModel

from .conftest import needs_charmhub_credentials


@needs_charmhub_credentials()
def test_charm_list_revisions(charm_client, charmhub_charm_name):
    revisions = charm_client.list_revisions(charmhub_charm_name)

    assert len(revisions) >= 1
    assert isinstance(revisions[-1], revisions_model.CharmRevisionModel)

    revision = cast(revisions_model.CharmRevisionModel, revisions[0])

    # Greater than or equal to in order to allow someone to replicate this
    # integration test themselves.
    assert revision.created_at >= datetime.datetime(
        2023, 4, 13, 16, 12, 55, tzinfo=datetime.timezone.utc
    )
    assert revision.size >= 400
    assert len(revision.sha3_384) == 96
    assert set(revision.sha3_384).issubset("0123456789abcdef")
    expected = revisions_model.CharmRevisionModel(
        created_at=revision.created_at,  # For replication purposes
        revision=1,
        sha3_384=revision.sha3_384,
        size=revision.size,
        status="released",
        bases=[CharmBaseModel(name="ubuntu", channel="22.04", architecture="amd64")],
        version="1",
    )
    assert revision == expected



#  -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*
#
#  Copyright 2023-2024 Canonical Ltd.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License version 3 as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Track model for Craft Store packages."""
from datetime import datetime
from typing import Annotated

import pydantic

from craft_store.models._base_model import MarshableModel


class TrackModel(MarshableModel):
    """A track that a package can be published on."""

    automatic_phasing_percentage: int | None = None
    created_at: Annotated[  # Prevents pydantic from setting UTC as "...Z"
        datetime,
        pydantic.WrapSerializer(
            lambda dt, _: dt.isoformat(), when_used="json-unless-none"
        ),
    ]
    name: str
    version_pattern: str | None = None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from unittest.mock import ANY, Mock, call, patch

import pytest
import requests
import urllib3
import urllib3.exceptions
from craft_store import HTTPClient, errors
from craft_store.http_client import _get_retry_value
from requests.exceptions import JSONDecodeError


def _fake_error_response(status_code, reason, json_raises=False):
    response = Mock()
    response.status_code = status_code
    response.ok = status_code == 200
    response.reason = reason
    if json_raises:
        response.json.side_effect = JSONDecodeError("foo", "doc", 0)
    return response


@pytest.fixture()
def session_mock():
    patched_session = patch("requests.Session", autospec=True)
    mocked_session = patched_session.start()
    mocked_session().request.return_value = _fake_error_response(200, "")
    yield mocked_session
    patched_session.stop()


@pytest.fixture()
def retry_mock():
    patched_retry = patch("craft_store.http_client.Retry", autospec=True)
    yield patched_retry.start()
    patched_retry.stop()


def test_session_defaults(session_mock, retry_mock):
    HTTPClient(user_agent="Secret Agent")

    assert [
        call("http://", ANY),
        call("https://", ANY),
    ] in session_mock().mount.mock_calls
    retry_mock.assert_called_once_with(
        total=8, backoff_factor=1, status_forcelist=[500, 502, 503, 504]
    )


def test_session_environment_values(monkeypatch, session_mock, retry_mock):
    monkeypatch.setenv("CRAFT_STORE_RETRIES", "20")
    monkeypatch.setenv("CRAFT_STORE_BACKOFF", "10")

    HTTPClient(user_agent="Secret Agent")

    assert [
        call("http://", ANY),
        call("https://", ANY),
    ] in session_mock().mount.mock_calls
    retry_mock.assert_called_once_with(
        total=20, backoff_factor=10, status_forcelist=[500, 502, 503, 504]
    )


@pytest.mark.parametrize("method", ["get", "post", "put"])
def test_methods(session_mock, method):
    client = HTTPClient(user_agent="Secret Agent")
    getattr(client, method)("https://foo.bar")

    assert session_mock().request.mock_calls == [
        call(
            method.upper(),
            "https://foo.bar",
            headers={"User-Agent": "Secret Agent"},
            params=None,
        )
    ]


scenarios = [
    {
        "expected_params": None,
        "expected_headers": {"User-Agent": "Secret Agent"},
        "expected_logger_debug_tail": "None and headers {'User-Agent': 'Secret Agent'}",
    },
    {
        "kwargs": {"headers": {"foo": "bar"}},
        "expected_params": None,
        "expected_headers": {"User-Agent": "Secret Agent", "foo": "bar"},
        "expected_logger_debug_tail": "None and headers {'foo': 'bar', 'User-Agent': 'Secret Agent'}",
    },
    {
        "kwargs": {"headers": {"Authorization": "bar"}},
        "expected_params": None,
        "expected_headers": {
            "User-Agent": "Secret Agent",
            "Authorization": "bar",
        },
        "expected_logger_debug_tail": (
            "None and headers {'Authorization': '<macaroon>', 'User-Agent': 'Secret Agent'}"
        ),
    },
    {
        "kwargs": {"headers": {"Macaroons": "bar"}},
        "expected_params": None,
        "expected_headers": {
            "User-Agent": "Secret Agent",
            "Macaroons": "bar",
        },
        "expected_logger_debug_tail": "None and headers {'Macaroons': '<macaroon>', 'User-Agent': 'Secret Agent'}",
    },
    {
        "kwargs": {"params": {"query": "bar"}},
        "expected_params": {"query": "bar"},
        "expected_headers": {
            "User-Agent": "Secret Agent",
        },
        "expected_logger_debug_tail": "{'query': 'bar'} and headers {'User-Agent': 'Secret Agent'}",
    },
    {
        "kwargs": {"params": {"query": "bar"}},
        "expected_params": {"query": "bar"},
        "expected_headers": {
            "User-Agent": "Secret Agent",
        },
        "expected_logger_debug_tail": "{'query': 'bar'} and headers {'User-Agent': 'Secret Agent'}",
    },
]


@pytest.mark.parametrize("scenario", scenarios)
def test_request(caplog, session_mock, scenario):
    caplog.set_level(logging.DEBUG)

    HTTPClient(user_agent="Secret Agent").request(
        "UPDATE",
        "https://foo.bar",
        **scenario.get("kwargs", {}),
    )

    assert session_mock().request.mock_calls == [
        call(
            "UPDATE",
            "https://foo.bar",
            headers=scenario["expected_headers"],
            params=scenario["expected_params"],
        )
    ]
    assert [
        "HTTP 'UPDATE' for 'https://foo.bar' with params "
        + scenario["expected_logger_debug_tail"]
    ] == [rec.message for rec in caplog.records]


def test_request_500(session_mock):
    fake_response = _fake_error_response(503, "cannot reach server", json_raises=True)
    session_mock().request.return_value = fake_response

    with pytest.raises(errors.StoreServerError) as store_error:
        HTTPClient(user_agent="Secret Agent").request(
            "GET",
            "https://foo.bar",
        )

    assert store_error.value.response == fake_response


def test_request_connection_error(session_mock):
    connection_pool = urllib3.connectionpool.ConnectionPool("https://foo.bar")
    connection_error = requests.exceptions.ConnectionError(
        urllib3.exceptions.MaxRetryError(pool=connection_pool, url="test-url")
    )
    session_mock().request.side_effect = connection_error

    with pytest.raises(errors.NetworkError) as network_error:
        HTTPClient(user_agent="Secret Agent").request(
            "GET",
            "https://foo.bar",
        )

    assert network_error.value.__cause__ == connection_error


def test_request_retry_error(session_mock):
    retry_error = requests.exceptions.RetryError()
    session_mock().request.side_effect = retry_error

    with pytest.raises(errors.NetworkError) as network_error:
        HTTPClient(user_agent="Secret Agent").request(
            "GET",
            "https://foo.bar",
        )

    assert network_error.value.__cause__ == retry_error


@pytest.mark.parametrize("environment_value", ["0", "10", "20000"])
def test_get_retry_value_environment_override(monkeypatch, caplog, environment_value):
    monkeypatch.setenv("FAKE_ENV", environment_value)
    caplog.set_level(logging.DEBUG)

    assert _get_retry_value("FAKE_ENV", 0) == int(environment_value)
    assert len(caplog.records) == 0


@pytest.mark.parametrize(
    ("environment_value", "default"), [("NaN", 10), ("NaN", 0.4), ("foo", 1)]
)
def test_get_retry_value_not_a_number_returns_default(
    monkeypatch, caplog, environment_value, default
):
    monkeypatch.setenv("FAKE_ENV", environment_value)
    caplog.set_level(logging.DEBUG)

    assert _get_retry_value("FAKE_ENV", default) == default
    assert [
        f"'FAKE_ENV' set to invalid value {environment_value!r}, setting to {default}."
    ] == [rec.message for rec in caplog.records]


@pytest.mark.parametrize(("environment_value", "default"), [("-1", 10), ("-1000", 0.4)])
def test_get_retry_value_negative_number_returns_default(
    monkeypatch, caplog, environment_value, default
):
    monkeypatch.setenv("FAKE_ENV", environment_value)
    caplog.set_level(logging.DEBUG)

    assert _get_retry_value("FAKE_ENV", default) == default
    int_value = int(environment_value)
    assert [
        f"'FAKE_ENV' set to non positive value {int_value!r}, setting to {default}."
    ] == [rec.message for rec in caplog.records]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022,2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""BaseModel with marshaling capabilities."""

from typing import Any

from pydantic import BaseModel, ConfigDict
from typing_extensions import Self


class MarshableModel(BaseModel):
    """A BaseModel that can be marshaled and unmarshaled."""

    model_config = ConfigDict(
        validate_assignment=True,
        frozen=True,
        alias_generator=lambda s: s.replace("_", "-"),
        populate_by_name=True,
    )

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> Self:
        """Create and populate a new ``MarshableModel`` from a dict.

        The unmarshal method validates entries in the input dictionary, populating
        the corresponding fields in the data object.

        :param data: The dictionary data to unmarshal.

        :return: The newly created object.

        :raise TypeError: If data is not a dictionary.
        """
        if not isinstance(data, dict):
            raise TypeError("part data is not a dictionary")

        return cls.model_validate(data)

    def marshal(self) -> dict[str, Any]:
        """Create a dictionary containing the part specification data.

        :return: The newly created dictionary.

        """
        return self.model_dump(mode="json", by_alias=True, exclude_unset=True)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023-2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Registered Names models for the Store."""
from typing import Any, Literal

import pydantic
from pydantic import AnyHttpUrl, Field

from ._base_model import MarshableModel
from .account_model import AccountModel
from .track_guardrail_model import TrackGuardrailModel
from .track_model import TrackModel


class MediaModel(MarshableModel):
    """Resource model for a media item attached to a registered name."""

    type: Literal["icon"]
    url: AnyHttpUrl


class RegisteredNameModel(MarshableModel):
    """Resource model for a registered name."""

    authority: str | None = None
    contact: str | None = None
    default_track: str | None = None
    description: str | None = None
    id: str
    links: dict[str, Any] = Field(default_factory=dict)
    media: list[MediaModel] = Field(default_factory=list)
    name: str | None = None
    private: bool
    publisher: AccountModel
    status: str
    store: str
    summary: str | None = None
    title: str | None = None
    track_guardrails: list[TrackGuardrailModel] = Field(default_factory=list)
    tracks: list[TrackModel] = Field(default_factory=list)
    type: str
    website: AnyHttpUrl | None = None

    @pydantic.field_serializer("website")
    def _serialize_website(self, website: AnyHttpUrl | None) -> str | None:
        if not website:
            return None
        return str(website)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from craft_store import endpoints


def test_charmhub():
    charmhub = endpoints.CHARMHUB

    assert charmhub.tokens == "/v1/tokens"
    assert charmhub.tokens_exchange == "/v1/tokens/exchange"
    assert charmhub.whoami == "/v1/tokens/whoami"
    assert charmhub.upload == "/unscanned-upload/"
    assert charmhub.get_token_request(
        permissions=["permission-foo", "permission-bar"],
        description="client description",
        ttl=1000,
    ) == {
        "permissions": ["permission-foo", "permission-bar"],
        "description": "client description",
        "ttl": 1000,
    }


def test_charmhub_channels():
    charmhub = endpoints.CHARMHUB

    assert charmhub.get_token_request(
        permissions=["permission-foo", "permission-bar"],
        description="client description",
        ttl=1000,
        channels=["stable", "track/edge"],
    ) == {
        "permissions": ["permission-foo", "permission-bar"],
        "description": "client description",
        "ttl": 1000,
        "channels": ["stable", "track/edge"],
    }


def test_charmhub_packages():
    charmhub = endpoints.CHARMHUB

    assert charmhub.get_token_request(
        permissions=["permission-foo", "permission-bar"],
        description="client description",
        ttl=1000,
        packages=[
            endpoints.Package("charm1", "charm"),
            endpoints.Package("bundle1", "bundle"),
        ],
    ) == {
        "permissions": ["permission-foo", "permission-bar"],
        "description": "client description",
        "ttl": 1000,
        "packages": [
            {"type": "charm", "name": "charm1"},
            {"type": "bundle", "name": "bundle1"},
        ],
    }


def test_charmhub_invalid_packages():
    charmhub = endpoints.CHARMHUB

    with pytest.raises(ValueError) as raised:  # noqa: PT011
        charmhub.get_token_request(
            permissions=["permission-foo", "permission-bar"],
            description="client description",
            ttl=1000,
            packages=[
                endpoints.Package("charm1", "snap"),
                endpoints.Package("bundle1", "rock"),
            ],
        )
    assert (
        str(raised.value) == "Package types ['snap', 'rock'] not in ['charm', 'bundle']"
    )


def test_charmhub_releases():
    charmhub = endpoints.CHARMHUB

    assert (
        charmhub.get_releases_endpoint("test-charm") == "/v1/charm/test-charm/releases"
    )


def test_charmhub_revisions():
    charmhub = endpoints.CHARMHUB

    assert (
        charmhub.get_revisions_endpoint("test-charm")
        == "/v1/charm/test-charm/revisions"
    )


def test_snap_store(expires):
    snap_store = endpoints.SNAP_STORE

    assert snap_store.tokens == "/api/v2/tokens"
    assert snap_store.tokens_exchange == "/api/v2/tokens/exchange"
    assert snap_store.whoami == "/api/v2/tokens/whoami"
    assert snap_store.upload == "/unscanned-upload/"
    assert snap_store.get_token_request(
        permissions=["permission-foo", "permission-bar"],
        description="client description",
        ttl=1000,
    ) == {
        "permissions": ["permission-foo", "permission-bar"],
        "description": "client description",
        "expires": expires(1000),
    }


def test_snap_store_channels(expires):
    snap_store = endpoints.SNAP_STORE

    assert snap_store.get_token_request(
        permissions=["permission-foo", "permission-bar"],
        description="client description",
        ttl=1000,
        channels=["stable", "track/edge"],
    ) == {
        "permissions": ["permission-foo", "permission-bar"],
        "description": "client description",
        "expires": expires(1000),
        "channels": ["stable", "track/edge"],
    }


def test_snap_store_packages(expires):
    snap_store = endpoints.SNAP_STORE

    assert snap_store.get_token_request(
        permissions=["permission-foo", "permission-bar"],
        description="client description",
        ttl=1000,
        packages=[
            endpoints.Package("snap1", "snap"),
            endpoints.Package("snap2", "snap"),
        ],
    ) == {
        "permissions": ["permission-foo", "permission-bar"],
        "description": "client description",
        "expires": expires(1000),
        "packages": [
            {"series": "16", "name": "snap1"},
            {"series": "16", "name": "snap2"},
        ],
    }


def test_snap_store_invalid_packages():
    snap_store = endpoints.SNAP_STORE

    with pytest.raises(ValueError) as raised:  # noqa: PT011
        snap_store.get_token_request(
            permissions=["permission-foo", "permission-bar"],
            description="client description",
            ttl=1000,
            packages=[
                endpoints.Package("snap1", "charm"),
                endpoints.Package("snap2", "rock"),
            ],
        )
    assert str(raised.value) == "Package types ['charm', 'rock'] not in ['snap']"


def test_snap_store_releases():
    snap_store = endpoints.SNAP_STORE

    with pytest.raises(NotImplementedError):
        snap_store.get_releases_endpoint("test-snap")


def test_snap_store_revisions():
    snap_store = endpoints.SNAP_STORE

    with pytest.raises(NotImplementedError):
        snap_store.get_revisions_endpoint("test-snap")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Generic Charm-related models.

These should only be used to import in other model files. External applications
should point to those imports in case the models change.
"""

from craft_store.models._base_model import MarshableModel


class CharmBaseModel(MarshableModel):
    """Base entries for the channel-map entry from the list_releases endpoint."""

    architecture: str
    channel: str
    name: str


class ResourceModel(MarshableModel):
    """Resource entries for the channel-map entry from the list_releases endpoint."""

    name: str
    revision: int | None = None
    type: str



#  -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
#  Copyright 2023 Canonical Ltd.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License version 3 as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for the store account model."""
import pytest
from craft_store.models.account_model import AccountModel

BASIC_ACCOUNT = {"id": "123"}
FULL_ACCOUNT = {
    "display-name": "Display Name",
    "email": "charmcrafters@lists.launchpad.net",
    "id": "abc123",
    "username": "usso-username",
    "validation": "unproven",
}


@pytest.mark.parametrize(
    ("json_dict", "expected"),
    [
        pytest.param(BASIC_ACCOUNT, AccountModel(id="123"), id="basic"),
        pytest.param(
            FULL_ACCOUNT,
            AccountModel(
                display_name="Display Name",  # pyright: ignore[reportCallIssue]
                # bug https://github.com/pydantic/pydantic/discussions/3986
                id="abc123",
                username="usso-username",
                validation="unproven",
                email="charmcrafters@lists.launchpad.net",
            ),
            id="fully-described",
        ),
    ],
)
def test_unmarshal(json_dict, expected):
    actual = AccountModel.unmarshal(json_dict)

    assert actual == expected


@pytest.mark.parametrize("payload", [BASIC_ACCOUNT, FULL_ACCOUNT])
def test_unmarshal_and_marshal(payload):
    assert AccountModel.unmarshal(payload).marshal() == payload



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""List Releases Models for Snap Store responses."""


from datetime import datetime

from ._base_model import MarshableModel
from ._common_list_releases_model import PackageModel, ProgressiveModel
from .revisions_model import SnapRevisionModel as RevisionModel


class ChannelMapModel(MarshableModel):
    """Model for the channel-map results from the list_releases endpoint."""

    architecture: str
    channel: str
    expiration_date: datetime | None = None
    progressive: ProgressiveModel
    revision: int
    when: datetime


class ListReleasesModel(MarshableModel):
    """Model for the list_releases endpoint."""

    channel_map: list[ChannelMapModel]
    package: PackageModel
    revisions: list[RevisionModel]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import cast

from craft_store.models import release_request_model


def test_release_model():
    model = release_request_model.ReleaseRequestModel(channel="edge", revision=1)

    assert model.channel == "edge"
    assert model.resources == []
    assert model.revision == 1


def test_release_model_to_close():
    model = release_request_model.ReleaseRequestModel(channel="edge", revision=None)

    assert model.channel == "edge"
    assert model.resources == []
    assert model.revision is None


def test_release_unmarshal_and_marshal():
    payload = {
        "channel": "stable",
        "revision": 2,
        "resources": [
            {
                "name": "resource-name-4",
                "revision": 4,
            },
            {
                "name": "resource-name-10",
                "revision": 10,
            },
        ],
    }

    model = cast(
        release_request_model.ReleaseRequestModel,
        release_request_model.ReleaseRequestModel.unmarshal(payload),
    )

    assert model.channel == "stable"
    assert model.revision == 2

    assert isinstance(model.resources, list)
    assert len(model.resources) == 2
    assert model.resources[0].name == "resource-name-4"
    assert model.resources[0].revision == 4
    assert model.resources[1].name == "resource-name-10"
    assert model.resources[1].revision == 10

    assert model.marshal() == payload



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022,2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from typing import Any
from unittest.mock import patch

import pytest


class FakeKeyring:
    """Fake Keyring Backend implementation for tests."""

    name = "Fake Keyring"

    def __init__(self) -> None:
        self.set_password_calls: list[tuple[Any, ...]] = []
        self.get_password_calls: list[tuple[Any, ...]] = []
        self.delete_password_calls: list[tuple[Any, ...]] = []
        self.password = None
        self.delete_error: Exception | None = None

    def set_password(self, *args) -> None:
        """Set the service password for username in memory."""
        self.set_password_calls.append(args)
        self.password = args[2]

    def get_password(self, *args):
        """Get the service password for username from memory."""
        self.get_password_calls.append(args)
        return self.password

    def delete_password(self, *args):
        """Delete the service password for username from memory."""
        self.delete_password_calls.append(args)
        if self.delete_error is not None:
            # https://www.logilab.org/ticket/3207
            raise self.delete_error


@pytest.fixture()
def keyring_set_keyring_mock():
    """Mock setting the keyring."""

    patched_keyring = patch("keyring.set_keyring", autospec=True)
    mocked_keyring = patched_keyring.start()
    yield mocked_keyring
    patched_keyring.stop()


@pytest.fixture()
def fake_keyring():
    return FakeKeyring()


@pytest.fixture(autouse=True)
def fake_keyring_get(fake_keyring, request):
    """Mock keyring and return a FakeKeyring."""
    if "disable_fake_keyring" in request.keywords:
        yield
    else:
        patched_keyring = patch("keyring.get_keyring")
        mocked_keyring = patched_keyring.start()
        mocked_keyring.return_value = fake_keyring
        yield mocked_keyring
        patched_keyring.stop()


@pytest.fixture()
def expires():
    """Mocks/freezes utcnow() in craft_store.endpoints module.

    Provides a function for creating expected iso formatted expires datetime
    values.
    """
    now = datetime.datetime.now(tz=datetime.timezone.utc)

    def offset_iso_dt(seconds=0):
        return (
            (now + datetime.timedelta(seconds=seconds))
            .replace(microsecond=0)
            .isoformat()
        )

    with patch("craft_store.endpoints.datetime", wraps=datetime.datetime) as dt_mock:
        dt_mock.utcnow.return_value = now
        yield offset_iso_dt


@pytest.fixture(params=[True, False], ids=["new_auth", "old_auth"])
def new_auth(request) -> bool:
    """
    Parametrized fixture representing either the new, type-based auth storage (True) or the old one (False).

    :see: base_client.wrap_credentials()
    """
    return request.param



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for list_releases."""
import hashlib
import time

import pytest
from craft_store.models.resource_revision_model import (
    CharmResourceRevision,
    CharmResourceType,
    RequestCharmResourceBase,
)

from .conftest import needs_charmhub_credentials


@needs_charmhub_credentials()
@pytest.mark.parametrize(
    "bases",
    [
        None,
        [],
        [
            RequestCharmResourceBase(
                name="ubuntu", channel="20.04", architectures=["all"]
            )
        ],
        [
            RequestCharmResourceBase(
                name="ubuntu", channel="22.04", architectures=["all"]
            ),
            RequestCharmResourceBase(
                name="ubuntu", channel="24.04", architectures=["all"]
            ),
        ],
    ],
)
def test_charm_push_resource(tmp_path, charm_client, charmhub_charm_name, bases):
    resource_name = "empty-file"
    path = tmp_path / resource_name
    path.unlink(missing_ok=True)
    path.touch()
    file_upload_id = charm_client.upload_file(filepath=path)
    file_status_url = charm_client.push_resource(
        charmhub_charm_name,
        resource_name,
        upload_id=file_upload_id,
        resource_type=CharmResourceType.FILE,
        bases=bases,
    )

    assert file_status_url.startswith(
        f"/v1/charm/{charmhub_charm_name}/revisions/review"
    )

    timeout = time.monotonic() + 120
    while True:
        file_status = charm_client.request(
            "GET", charm_client._base_url + file_status_url
        ).json()["revisions"][0]
        if file_status["status"] in ("approved", "rejected"):
            break
        if time.monotonic() > timeout:
            raise TimeoutError(
                "Zero-byte file upload was not approved or rejected after 120s",
                file_status,
            )
        time.sleep(2)
    assert file_status["revision"] is not None
    file_revisions = charm_client.list_resource_revisions(
        name=charmhub_charm_name, resource_name=resource_name
    )
    for revision in file_revisions:
        if revision.revision == file_status["revision"]:
            break
    else:
        raise ValueError(
            "Zero-byte file revision from status URL does not appear in revisions."
        )
    assert revision.size == 0
    assert revision.sha3_384 == hashlib.sha3_384(b"").hexdigest()
    assert "all" in revision.bases[0].architectures

    assert len(file_revisions) >= 1
    assert isinstance(file_revisions[-1], CharmResourceRevision)



#  -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
#  Copyright 2023-2024 Canonical Ltd.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License version 3 as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for the store trock model."""
from datetime import datetime

import pytest
from craft_store.models.track_model import TrackModel

BASIC_TRACK = {"created-at": "2023-03-28T18:50:44+00:00", "name": "1.0/stable"}
FULL_TRACK = {
    "automatic-phasing-percentage": "10",
    "created-at": "2023-03-28T18:50:44+00:00",
    "name": "1.0/stable",
    "version-pattern": r"^\d\.\d/",
}


@pytest.mark.parametrize(
    "json_dict",
    [
        pytest.param(BASIC_TRACK, id="basic"),
        pytest.param(FULL_TRACK, id="fully-described"),
    ],
)
def test_unmarshal(check, json_dict):
    actual = TrackModel.unmarshal(json_dict)

    check.equal(actual.name, json_dict["name"])
    check.equal(actual.created_at, datetime.fromisoformat(json_dict["created-at"]))
    pct = json_dict.get("automatic-phasing-percentage")
    if isinstance(pct, str):
        pct = int(pct)
    check.equal(actual.automatic_phasing_percentage, pct)
    check.equal(actual.version_pattern, json_dict.get("version-pattern"))


@pytest.mark.parametrize("payload", [BASIC_TRACK, FULL_TRACK])
def test_unmarshal_and_marshal(payload, check):
    marshalled = TrackModel.unmarshal(payload).marshal()

    check.equal(marshalled["created-at"], payload["created-at"])
    check.equal(marshalled["name"], payload["name"])
    check.equal(
        "automatic-phasing-percentage" in marshalled,
        "automatic-phasing-percentage" in payload,
    )
    if "automatic-phasing-percentage" in payload:
        phasing_percentage = int(payload.get("automatic-phasing-percentage"))
    else:
        phasing_percentage = None
    check.equal(marshalled.get("automatic-phasing-percentage"), phasing_percentage)
    check.equal("version-pattern" in marshalled, "version-pattern" in payload)
    check.equal(marshalled.get("version-pattern"), payload.get("version-pattern"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent
from unittest import mock

import pytest
import requests
import urllib3
import urllib3.exceptions
from craft_store import errors
from requests.exceptions import JSONDecodeError


def _fake_error_response(status_code, reason, json=None):
    response = mock.Mock()
    response.status_code = status_code
    response.reason = reason
    if json is None:
        response.json.side_effect = JSONDecodeError("foo", "doc", 0)
    else:
        response.json = mock.Mock(return_value=json)

    return response


scenarios = (
    {
        "exception_class": errors.NetworkError,
        "args": [requests.exceptions.ConnectionError("bad error")],
        "expected_message": "bad error",
    },
    {
        "exception_class": errors.NetworkError,
        "args": [
            requests.exceptions.ConnectionError(
                urllib3.exceptions.MaxRetryError(
                    pool=urllib3.connectionpool.ConnectionPool("https://foo.bar"),
                    url="test-url",
                )
            )
        ],
        "expected_message": "Maximum retries exceeded trying to reach the store.",
    },
    {
        "exception_class": errors.StoreServerError,
        "args": [_fake_error_response(500, "internal server error")],
        "expected_message": "Issue encountered while processing your request: [500] internal server error.",
    },
    {
        "exception_class": errors.StoreServerError,
        "args": [_fake_error_response(501, "not implemented")],
        "expected_message": "Issue encountered while processing your request: [501] not implemented.",
    },
    {
        "exception_class": errors.NoKeyringError,
        "args": [],
        "expected_message": "No keyring found to store or retrieve credentials from.",
    },
    {
        "exception_class": errors.KeyringUnlockError,
        "args": [],
        "expected_message": "Failed to unlock the keyring.",
    },
    {
        "exception_class": errors.CredentialsAlreadyAvailable,
        "args": ["mycraft", "my.host.com"],
        "expected_message": "Credentials found for 'mycraft' on 'my.host.com'.",
    },
    {
        "exception_class": errors.CredentialsUnavailable,
        "args": ["mycraft", "my.host.com"],
        "expected_message": "No credentials found for 'mycraft' on 'my.host.com'.",
    },
    {
        "exception_class": errors.CredentialsNotParseable,
        "args": [],
        "expected_message": "Credentials could not be parsed. Expected base64 encoded credentials.",
    },
    {
        "exception_class": errors.CandidTokenTimeoutError,
        "args": ["https://foo.bar"],
        "expected_message": "Timed out waiting for token response from 'https://foo.bar'.",
    },
    {
        "exception_class": errors.CandidTokenKindError,
        "args": ["https://foo.bar"],
        "expected_message": "Empty token kind returned from 'https://foo.bar'.",
    },
    {
        "exception_class": errors.CandidTokenValueError,
        "args": ["https://foo.bar"],
        "expected_message": "Empty token value returned from 'https://foo.bar'.",
    },
)


@pytest.mark.parametrize("scenario", scenarios)
def test_error_formatting(scenario):
    assert (
        str(scenario["exception_class"](*scenario["args"]))
        == scenario["expected_message"]
    )


error_lists = [
    {
        "error_list": [
            {"code": "resource-not-found", "message": "could not find resource"},
        ],
        "expected": dedent(
            """\
            Store operation failed:
            - resource-not-found: could not find resource"""
        ),
    },
    {
        "error_list": [
            {"code": "resource-not-found", "message": "could not find resource"},
            {
                "code": "resource-doubly-not-found",
                "message": "could not find resource when trying harder",
            },
        ],
        "expected": dedent(
            """\
            Store operation failed:
            - resource-not-found: could not find resource
            - resource-doubly-not-found: could not find resource when trying harder"""
        ),
    },
]


@pytest.mark.parametrize("error_list_key", ["error-list", "error_list"])
@pytest.mark.parametrize("error_list", error_lists)
def test_store_error_list(error_list_key, error_list):
    response = _fake_error_response(
        404, "resource-not-found", json={error_list_key: error_list["error_list"]}
    )
    assert str(errors.StoreServerError(response)) == error_list["expected"]


@pytest.mark.parametrize("missing", ["code", "message"])
@pytest.mark.parametrize("error_list_key", ["error-list", "error_list"])
def test_store_error_list_missing_element(missing, error_list_key):
    error_list = [
        {"code": "resource-not-found", "message": "could not find resource"},
    ]
    error_list[0].pop(missing)
    response = _fake_error_response(
        404, "resource-not-found", json={error_list_key: error_list}
    )
    assert (
        str(errors.StoreServerError(response))
        == "Issue encountered while processing your request: [404] resource-not-found."
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
from unittest.mock import ANY

import keyring
import keyring.backends.fail
import keyring.errors
import pytest
from craft_store import errors
from craft_store.auth import Auth, FileKeyring, MemoryKeyring
from keyring.backends import SecretService


def test_set_credentials(caplog, fake_keyring):
    auth = Auth("fakeclient", "fakestore.com")

    auth.set_credentials("{'password': 'secret'}")

    assert fake_keyring.set_password_calls == [
        ("fakeclient", "fakestore.com", "eydwYXNzd29yZCc6ICdzZWNyZXQnfQ=="),
    ]
    assert caplog.records == []


def test_set_credentials_log_debug(caplog, fake_keyring):
    caplog.set_level(logging.DEBUG)
    auth = Auth("fakeclient", "fakestore.com")

    auth.set_credentials("{'password': 'secret'}")

    assert fake_keyring.set_password_calls == [
        ("fakeclient", "fakestore.com", "eydwYXNzd29yZCc6ICdzZWNyZXQnfQ=="),
    ]
    assert [
        "Storing credentials for 'fakeclient' on 'fakestore.com' in keyring 'Fake Keyring'."
    ] == [rec.message for rec in caplog.records]


@pytest.mark.usefixtures("fake_keyring")
def test_double_set_credentials_fails():
    auth = Auth("fakeclient", "fakestore.com")

    auth.set_credentials("{'password': 'secret'}")

    with pytest.raises(errors.CredentialsAlreadyAvailable):
        auth.set_credentials("{'password': 'secret'}")


def test_double_set_credentials_force(fake_keyring):
    auth = Auth("fakeclient", "fakestore.com")

    auth.set_credentials("{'password': 'secret'}")

    auth.set_credentials("{'password': 'secret2'}", force=True)

    assert fake_keyring.set_password_calls == [
        ("fakeclient", "fakestore.com", "eydwYXNzd29yZCc6ICdzZWNyZXQnfQ=="),
        ("fakeclient", "fakestore.com", "eydwYXNzd29yZCc6ICdzZWNyZXQyJ30="),
    ]


def test_get_credentials(caplog, fake_keyring):
    fake_keyring.password = "eydwYXNzd29yZCc6ICdzZWNyZXQnfQ=="  # noqa: S105

    auth = Auth("fakeclient", "fakestore.com")

    assert auth.get_credentials() == "{'password': 'secret'}"
    assert fake_keyring.get_password_calls == [("fakeclient", "fakestore.com")]
    assert caplog.records == []


def test_get_credentials_log_debug(caplog, fake_keyring):
    caplog.set_level(logging.DEBUG)
    fake_keyring.password = "eydwYXNzd29yZCc6ICdzZWNyZXQnfQ=="  # noqa: S105

    auth = Auth("fakeclient", "fakestore.com")

    assert auth.get_credentials() == "{'password': 'secret'}"
    assert fake_keyring.get_password_calls == [("fakeclient", "fakestore.com")]
    assert [
        "Retrieving credentials for 'fakeclient' on 'fakestore.com' from keyring 'Fake Keyring'."
    ] == [rec.message for rec in caplog.records]


def test_get_credentials_no_credentials_in_keyring(caplog, fake_keyring):
    auth = Auth("fakeclient", "fakestore.com")

    with pytest.raises(errors.CredentialsUnavailable):
        auth.get_credentials()

    assert fake_keyring.get_password_calls == [("fakeclient", "fakestore.com")]
    assert caplog.records == []


def test_del_credentials(caplog, fake_keyring):
    fake_keyring.password = "eydwYXNzd29yZCc6ICdzZWNyZXQnfQ=="  # noqa: S105

    auth = Auth("fakeclient", "fakestore.com")

    auth.del_credentials()

    assert fake_keyring.delete_password_calls == [("fakeclient", "fakestore.com")]
    assert caplog.records == []


def test_del_credentials_log_debug(caplog, fake_keyring):
    caplog.set_level(logging.DEBUG)
    fake_keyring.password = "eydwYXNzd29yZCc6ICdzZWNyZXQnfQ=="  # noqa: S105

    auth = Auth("fakeclient", "fakestore.com")

    auth.del_credentials()

    assert fake_keyring.delete_password_calls == [("fakeclient", "fakestore.com")]
    assert [
        "Retrieving credentials for 'fakeclient' on 'fakestore.com' from keyring 'Fake Keyring'.",
        "Deleting credentials for 'fakeclient' on 'fakestore.com' from keyring 'Fake Keyring'.",
    ] == [rec.message for rec in caplog.records]


def test_del_credentials_delete_error_in_keyring(caplog, fake_keyring):
    fake_keyring.delete_error = keyring.errors.PasswordDeleteError()
    fake_keyring.password = "eydwYXNzd29yZCc6ICdzZWNyZXQnfQ=="  # noqa: S105

    auth = Auth("fakeclient", "fakestore.com")

    with pytest.raises(keyring.errors.PasswordDeleteError):
        auth.del_credentials()

    assert fake_keyring.delete_password_calls == [("fakeclient", "fakestore.com")]
    assert caplog.records == []


def test_del_credentials_gets_no_credential(caplog, fake_keyring):
    caplog.set_level(logging.DEBUG)

    auth = Auth("fakeclient", "fakestore.com")

    with pytest.raises(errors.CredentialsUnavailable):
        auth.del_credentials()

    assert fake_keyring.get_password_calls == [("fakeclient", "fakestore.com")]
    assert [
        "Retrieving credentials for 'fakeclient' on 'fakestore.com' from keyring 'Fake Keyring'.",
        "Credentials not found in the keyring 'Fake Keyring'",
    ] == [rec.message for rec in caplog.records]


def test_credentials_not_parseable_error(monkeypatch):
    monkeypatch.setenv("FAKE_ENV", "12345")

    with pytest.raises(errors.CredentialsNotParseable):
        Auth("fakeclient", "fakestore.com", environment_auth="FAKE_ENV")


def test_environment_set(monkeypatch, fake_keyring, keyring_set_keyring_mock):
    monkeypatch.setenv("FAKE_ENV", "c2VjcmV0LWtleXM=")

    Auth("fakeclient", "fakestore.com", environment_auth="FAKE_ENV")

    assert keyring_set_keyring_mock.mock_calls == [ANY]
    assert fake_keyring.set_password_calls == [
        ("fakeclient", "fakestore.com", "c2VjcmV0LWtleXM=")
    ]


def test_ensure_no_credentials_unlock_error(fake_keyring, mocker):
    mocker.patch.object(
        fake_keyring, "get_password", side_effect=errors.KeyringUnlockError
    )

    auth = Auth("fakeclient", "fakestore.com")

    with pytest.raises(errors.KeyringUnlockError):
        auth.ensure_no_credentials()


@pytest.mark.disable_fake_keyring()
def test_ephemeral_set_memory_keyring():
    auth = Auth("fakeclient", "fakestore.com", ephemeral=True)

    assert isinstance(auth._keyring, MemoryKeyring)


def test_no_keyring_get(fake_keyring_get):
    fake_keyring_get.return_value = keyring.backends.fail.Keyring()

    with pytest.raises(errors.NoKeyringError):
        Auth("fakeclient", "fakestore.com")


def test_memory_keyring_set_get():
    k = MemoryKeyring()
    k.set_password("my-service", "my-user", "my-password")

    assert k.get_password("my-service", "my-user") == "my-password"


def test_memory_keyring_get_empty():
    k = MemoryKeyring()

    assert k.get_password("my-service", "my-user") is None


def test_memory_keyring_set_delete():
    k = MemoryKeyring()
    k.set_password("my-service", "my-user", "my-password")

    assert k.get_password("my-service", "my-user") == "my-password"

    k.delete_password("my-service", "my-user")

    assert k.get_password("my-service", "my-user") is None


def test_memory_keyring_delete_empty():
    k = MemoryKeyring()

    with pytest.raises(keyring.errors.PasswordDeleteError):
        k.delete_password("my-service", "my-user")


@pytest.fixture(autouse=True)
def _fake_basedirectory(mocker, tmp_path):
    mocker.patch("craft_store.auth.BaseDirectory.save_data_path", return_value=tmp_path)


def test_file_keyring_set_get():
    k = FileKeyring("test-app")
    k.set_password("my-service", "my-user", "my-password")

    assert k.get_password("my-service", "my-user") == "my-password"


def test_file_keyring_get_empty():
    k = FileKeyring("test-app")

    assert k.get_password("my-service", "my-user") is None


def test_file_keyring_set_delete():
    k = FileKeyring("test-app")
    k.set_password("my-service", "my-user", "my-password")

    assert k.get_password("my-service", "my-user") == "my-password"

    k.delete_password("my-service", "my-user")

    assert k.get_password("my-service", "my-user") is None


def test_file_keyring_delete_empty():
    k = FileKeyring("test-app")

    with pytest.raises(keyring.errors.PasswordDeleteError):
        k.delete_password("my-service", "my-user")


def test_file_keyring_storage_path(tmp_path):
    """Ensure the mock is used."""
    k = FileKeyring("test-app")

    assert k.credentials_file == tmp_path / "credentials.json"


test_exceptions: list[type[Exception]] = [keyring.errors.InitError]
if sys.platform == "linux":
    from secretstorage.exceptions import SecretServiceNotAvailableException

    test_exceptions.append(SecretServiceNotAvailableException)


@pytest.mark.disable_fake_keyring()
@pytest.mark.parametrize("exception", test_exceptions)
def test_secretservice_file_fallsback(mocker, exception):
    # At one point in the code we run keyring.set_backend, there is no
    # elegant way to reset this in the library.
    keyring.set_keyring(SecretService.Keyring())
    mocker.patch(
        "keyring.backends.SecretService.Keyring.get_preferred_collection",
        side_effect=exception,
    )
    auth = Auth(application_name="test-app", host="foo")

    assert type(auth._keyring) == FileKeyring


@pytest.mark.disable_fake_keyring()
def test_secretservice_works(mocker):
    # At one point in the code we run keyring.set_backend, there is no
    # elegant way to reset this in the library.
    keyring.set_keyring(SecretService.Keyring())
    mocker.patch(
        "keyring.backends.SecretService.Keyring.get_preferred_collection",
        return_value=None,
    )
    auth = Auth(application_name="test-app", host="foo")

    assert type(auth._keyring) == SecretService.Keyring



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023-2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for resource revision models."""
import pydantic
import pytest
from craft_store.models import CharmResourceRevisionUpdateRequest


@pytest.mark.parametrize(
    ("request_dict", "match"),
    [
        ({"revision": 1}, r"bases[:\s]+Field required"),
        (
            {"revision": 1, "bases": []},
            r"bases[:\s]+List should have at least 1 item",
        ),
        (
            {"revision": 1, "bases": [{"architectures": ["all", "all"]}]},
            r"bases.0.architectures[:\s]+Value error, Duplicate values in list:",
        ),
        (
            {"revision": 1, "bases": [{"architectures": []}]},
            r"bases.0.architectures[:\s]+List should have at least 1 item",
        ),
    ],
)
def test_charmresourcerevisionupdaterequest_invalid_bases(request_dict, match):
    with pytest.raises(pydantic.ValidationError, match=match):
        CharmResourceRevisionUpdateRequest.unmarshal(request_dict)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import pytest
from craft_store import creds, errors


@pytest.fixture()
def stored_candid_creds(new_auth: bool) -> str:
    """Fixture that generates Candid credentials in the format read from storage.

    The credentials will either be a raw string (if `new_auth` is False) or a JSON'ed
    dict encoding the credentials type (if `new_auth` is True).
    """
    candid_creds = "fake candid creds"
    if new_auth:
        return json.dumps({"t": "macaroon", "v": candid_creds})

    return candid_creds


@pytest.fixture()
def stored_u1_creds(new_auth: bool) -> str:
    """Fixture that generates Ubuntu One credentials in the format read from storage.

    The stored credentials are always a JSON'ed dict. The dict itself is either in the
    "old" format listing the root and discharge macaroons (if `new_auth` is False) or
    in the "new" format listing the credentials type *and* the macaroons (if `new_auth`
    is True).
    """
    u1_creds = {"r": "fake root macaroon", "d": "fake discharge macaroon"}
    if new_auth:
        return json.dumps({"t": "u1-macaroon", "v": u1_creds})
    return json.dumps(u1_creds)


def test_candid_creds_unmarshal(stored_candid_creds):
    """Test that we can parse both types of stored Candid creds."""
    candid_creds = creds.unmarshal_candid_credentials(stored_candid_creds)
    assert candid_creds == "fake candid creds"


def test_candid_creds_marshal():
    marshalled = creds.marshal_candid_credentials("fake candid creds")
    loaded = json.loads(marshalled)

    assert len(loaded) == 2
    assert loaded["t"] == "macaroon"
    assert loaded["v"] == "fake candid creds"


def test_u1_creds_unmarshal(stored_u1_creds):
    """Test that we can parse both types of stored Ubuntu One creds."""
    u1_creds = creds.unmarshal_u1_credentials(stored_u1_creds)
    assert u1_creds.root == "fake root macaroon"
    assert u1_creds.discharge == "fake discharge macaroon"


def test_u1_creds_marshal():
    macaroons = creds.UbuntuOneMacaroons(r="fake root", d="fake discharge")
    marshalled = creds.marshal_u1_credentials(macaroons)
    loaded = json.loads(marshalled)

    assert len(loaded) == 2
    assert loaded["t"] == "u1-macaroon"
    assert len(loaded["v"]) == 2
    assert loaded["v"]["r"] == "fake root"
    assert loaded["v"]["d"] == "fake discharge"


def test_u1_creds_unmarshal_failure():
    with pytest.raises(errors.CredentialsNotParseable):
        creds.unmarshal_u1_credentials("not a valid json string")


def test_mixed_creds():
    """Test the case of trying to load the wrong type of credentials."""
    macaroons = creds.UbuntuOneMacaroons(r="fake root", d="fake discharge")
    stored_u1 = creds.marshal_u1_credentials(macaroons)

    stored_candid = creds.marshal_candid_credentials("fake candid")

    # Try to load Ubuntu One as Candid:
    with pytest.raises(errors.CredentialsNotParseable):
        creds.unmarshal_candid_credentials(stored_u1)

    # Try to load Candid as Ubuntu One:
    with pytest.raises(errors.CredentialsNotParseable):
        creds.unmarshal_u1_credentials(stored_candid)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime
from typing import cast

import pytest
from craft_store.models import charm_list_releases_model

from .conftest import needs_charmhub_credentials

pytestmark = pytest.mark.timeout(10)  # Timeout if any test takes over 10 sec.


@needs_charmhub_credentials()
def test_charm_get_list_releases(charm_client, charmhub_charm_name):
    """Test list releases for a given charm.

    If you need to create this for yourself you can replicate the
    Charmcraft "resources" spread test:
    https://github.com/canonical/charmcraft/blob/main/tests/spread/store/resources/task.yaml
    but with only the Docker image, not the example file.

    Only upload a single revision, and release that to the edge channel.

    Set the charm name to something not already registered in the staging store.
    Set the environment variable CRAFT_STORE_TEST_CHARM to the name of the
    charm you registered before running this test. This will ensure that the
    test will run against your charm.
    """
    model = cast(
        charm_list_releases_model.ListReleasesModel,
        charm_client.get_list_releases(name=charmhub_charm_name),
    )

    assert len(model.channel_map) == 1
    assert model.channel_map[0].base.architecture == "amd64"
    assert model.channel_map[0].base.channel == "22.04"
    assert model.channel_map[0].base.name == "ubuntu"
    assert model.channel_map[0].channel == "latest/edge"
    assert model.channel_map[0].expiration_date is None
    assert model.channel_map[0].progressive.paused is None
    assert model.channel_map[0].progressive.percentage is None
    assert model.channel_map[0].resources == [
        charm_list_releases_model.ResourceModel(
            name="example-image", revision=1, type="oci-image"
        )
    ]
    assert model.channel_map[0].revision == 1
    # Greater than or equal to in order to allow someone to replicate this
    # integration test themselves.
    assert model.channel_map[0].when >= datetime.datetime(
        2023, 4, 13, 16, 12, 55, tzinfo=datetime.timezone.utc
    )

    assert len(model.package.channels) == 4
    assert model.package.channels[0].branch is None
    assert model.package.channels[0].fallback is None
    assert model.package.channels[0].name == "latest/stable"
    assert model.package.channels[0].risk == "stable"
    assert model.package.channels[0].track == "latest"
    assert model.package.channels[1].branch is None
    assert model.package.channels[1].fallback == "latest/stable"
    assert model.package.channels[1].name == "latest/candidate"
    assert model.package.channels[1].risk == "candidate"
    assert model.package.channels[1].track == "latest"
    assert model.package.channels[2].branch is None
    assert model.package.channels[2].fallback == "latest/candidate"
    assert model.package.channels[2].name == "latest/beta"
    assert model.package.channels[2].risk == "beta"
    assert model.package.channels[2].track == "latest"
    assert model.package.channels[3].branch is None
    assert model.package.channels[3].fallback == "latest/beta"
    assert model.package.channels[3].name == "latest/edge"
    assert model.package.channels[3].risk == "edge"
    assert model.package.channels[3].track == "latest"

    assert len(model.revisions) == 1
    assert len(model.revisions[0].bases) == 1
    assert model.revisions[0].bases[0].architecture == "amd64"
    assert model.revisions[0].bases[0].channel == "22.04"
    assert model.revisions[0].bases[0].name == "ubuntu"
    # No timezone information returned from Charmhub.
    # Greater than or equal to in order to allow someone to replicate this
    # integration test themselves.
    assert model.revisions[0].created_at >= datetime.datetime(
        2023, 4, 13, 16, 9, 55, 19472
    )
    assert model.revisions[0].revision == 1
    assert (
        model.revisions[0].sha3_384
        == "9c1368ba01e30aff43c3372ed61b7cdfc3330b3a3044d887964ccd8100fe2ea59f13409a70596107f981bd09cc9d9b21"
    )
    assert model.revisions[0].size == 6119029
    assert model.revisions[0].status == "released"
    assert model.revisions[0].version == "1"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Full workflow tests for charms."""
import datetime
import hashlib
import time

from craft_store.models import revisions_model
from craft_store.models.resource_revision_model import (
    CharmResourceRevisionUpdateRequest,
    CharmResourceType,
    RequestCharmResourceBase,
    ResponseCharmResourceBase,
)

from tests.integration.conftest import needs_charmhub_credentials


@needs_charmhub_credentials()
# This is intentionally long since it goes through a full workflow
def test_full_charm_workflow(  # noqa: PLR0912, PLR0915
    tmp_path, charm_client, charmhub_charm_name, fake_charms
):
    """A full workflow test for uploading a charm.

    Steps include:
    1. Check if charm name is registered.
       1.1. Register if necessary.
    2. Upload one or more charm revisions
    3. Upload one or more OCI images (must already exist on disk) - TODO
    4. Upload a fresh file
    5. Upload an empty file
    6. Modify the bases for a resource.
    """
    # 1.
    charmhub_charm_name += "-workflow"
    registered_names = charm_client.list_registered_names(include_collaborations=True)
    for name in registered_names:
        if name.name == charmhub_charm_name:
            break
    else:
        charm_client.register_name(charmhub_charm_name, entity_type="charm")

    # 2.
    upload_ids = [charm_client.upload_file(filepath=charm) for charm in fake_charms]
    revision_status_urls = [
        charm_client.notify_revision(
            name=charmhub_charm_name,
            revision_request=revisions_model.RevisionsRequestModel(upload_id=upload_id),
        ).status_url
        for upload_id in upload_ids
    ]

    timeout = time.monotonic() + 120
    while True:
        revision_statuses = [
            # Replace this with list_upload_reviews when working on
            # https://github.com/canonical/craft-store/issues/138
            charm_client.request("GET", charm_client._base_url + status_url).json()
            for status_url in revision_status_urls
        ]
        for status in revision_statuses:
            if status["revisions"][0]["status"] not in ("approved", "rejected"):
                time.sleep(2)  # Checking every 2 seconds seems reasonable
                break  # out of the for loop
        else:
            break  # out of the while loop
        if time.monotonic() >= timeout:
            raise TimeoutError(
                "Waited over 120 seconds, charm uploads still neither approved nor rejected",
                revision_statuses,
            )

    revisions_numbers = [
        status["revisions"][0]["revision"] for status in revision_statuses
    ]

    if None in revisions_numbers:
        raise AssertionError(
            "One or more revisions did not get a number", revision_statuses
        )

    # 3. Rocks!
    # TODO: This requires the docker registry and stuff

    # Prep for file upload stuff
    arch_dependent_base = RequestCharmResourceBase(architectures=["amd64"])
    neutral_base = RequestCharmResourceBase(architectures=["all"])

    # 4. Upload a fresh file.
    fresh_file = tmp_path / "fresh_file"
    file_contents = datetime.datetime.now(tz=datetime.timezone.utc).isoformat().encode()
    fresh_file.write_bytes(file_contents)
    file_upload_id = charm_client.upload_file(filepath=fresh_file)
    file_status_url = charm_client.push_resource(
        charmhub_charm_name,
        "my-file",
        upload_id=file_upload_id,
        resource_type=CharmResourceType.FILE,
        bases=[arch_dependent_base],
    )  # Response is a list_upload_reviews path + query parameter.
    timeout = time.monotonic() + 120
    while True:
        file_status = charm_client.request(
            "GET", charm_client._base_url + file_status_url
        ).json()["revisions"][0]
        if file_status["status"] in ("approved", "rejected"):
            break
        if time.monotonic() > timeout:
            raise TimeoutError(
                "File upload was not approved or rejected after 120s", file_status
            )
        time.sleep(2)
    assert file_status["revision"] is not None
    file_revisions = charm_client.list_resource_revisions(
        name=charmhub_charm_name, resource_name="my-file"
    )
    for file_revision in file_revisions:
        if file_revision.revision == file_status["revision"]:
            break
    else:
        raise ValueError("File revision from status URL does not appear in revisions.")

    file_sha256 = hashlib.sha256(file_contents).hexdigest()
    assert file_revision.size == len(
        file_contents
    ), "Uploaded file size does not match file."
    assert (
        file_revision.sha256 == file_sha256
    ), "Uploaded file hash does not match file."
    assert file_revision.bases == [arch_dependent_base]

    # 5. Upload a zero-byte file (probably a repeat)
    zero_byte_file = tmp_path / "zero_bytes"
    zero_byte_file.touch()
    file_upload_id = charm_client.upload_file(filepath=zero_byte_file)
    file_status_url = charm_client.push_resource(
        charmhub_charm_name,
        "my-file",
        upload_id=file_upload_id,
        resource_type=CharmResourceType.FILE,
        bases=[arch_dependent_base],
    )
    timeout = time.monotonic() + 120
    while True:
        file_status = charm_client.request(
            "GET", charm_client._base_url + file_status_url
        ).json()["revisions"][0]
        if file_status["status"] in ("approved", "rejected"):
            break
        if time.monotonic() > timeout:
            raise TimeoutError(
                "Zero-byte file upload was not approved or rejected after 120s",
                file_status,
            )
        time.sleep(2)
    assert file_status["revision"] is not None
    file_revisions = charm_client.list_resource_revisions(
        name=charmhub_charm_name, resource_name="my-file"
    )
    for zb_file_revision in file_revisions:
        if zb_file_revision.revision == file_status["revision"]:
            break
    else:
        raise ValueError(
            "Zero-byte file revision from status URL does not appear in revisions."
        )
    assert zb_file_revision.size == 0
    assert zb_file_revision.sha3_384 == hashlib.sha3_384(b"").hexdigest()
    assert "amd64" in zb_file_revision.bases[0].architectures

    # 6. Modify bases for the files.
    assert (
        charm_client.update_resource_revisions(
            CharmResourceRevisionUpdateRequest(
                revision=file_revision.revision,
                bases=[arch_dependent_base, neutral_base],
            ),
            CharmResourceRevisionUpdateRequest(
                revision=zb_file_revision.revision,
                bases=[arch_dependent_base, neutral_base],
            ),
            name=charmhub_charm_name,
            resource_name="my-file",
        )
        == 2
    )
    file_revisions = charm_client.list_resource_revisions(
        name=charmhub_charm_name, resource_name="my-file"
    )
    new_zb_revision = None
    new_revision = None
    for rev in file_revisions:
        if rev.revision == zb_file_revision.revision:
            new_zb_revision = rev
        if rev.revision == file_revision.revision:
            new_revision = rev
        if new_revision and new_zb_revision:
            break
    assert new_revision is not None, "File revision not found in resource revisions"
    assert (
        new_zb_revision is not None
    ), "Zero-byte file revision not found in resource revisions"
    combined_base = ResponseCharmResourceBase(
        name="all", channel="all", architectures=["amd64", "all"]
    )
    assert new_revision.bases == [combined_base]
    assert new_zb_revision.bases == [combined_base]

    charm_client.update_resource_revision(
        charmhub_charm_name,
        "my-file",
        revision=zb_file_revision.revision,
        bases=[neutral_base],
    )
    file_revisions = charm_client.list_resource_revisions(
        name=charmhub_charm_name, resource_name="my-file"
    )
    for rev in file_revisions:
        if rev.revision == zb_file_revision.revision:
            break
    else:
        raise ValueError("No file with the appropriate revision found")
    assert rev.bases == [neutral_base]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import keyring
import keyring.backend
import keyring.errors
import pytest
from craft_store import errors
from craft_store.auth import Auth, MemoryKeyring

pytestmark = pytest.mark.timeout(10)  # Timeout if any test takes over 10 sec.


@pytest.fixture()
def _test_keyring():
    """In memory keyring backend for testing."""
    current_keyring = keyring.get_keyring()
    keyring.set_keyring(MemoryKeyring())
    yield
    keyring.set_keyring(current_keyring)


@pytest.mark.usefixtures("_test_keyring")
def test_auth():
    auth = Auth("fakecraft", "fakestore.com")

    auth.set_credentials("foo")
    assert auth.get_credentials() == "foo"
    auth.del_credentials()

    with pytest.raises(errors.CredentialsUnavailable) as error:
        auth.get_credentials()

    assert str(error.value) == (
        "No credentials found for 'fakecraft' on 'fakestore.com'."
    )

    with pytest.raises(errors.CredentialsUnavailable) as error:
        auth.del_credentials()

    assert str(error.value) == (
        "No credentials found for 'fakecraft' on 'fakestore.com'."
    )


def test_auth_from_environment(monkeypatch):
    monkeypatch.setenv("CREDENTIALS", "c2VjcmV0LWtleXM=")

    auth = Auth("fakecraft", "fakestore.com", environment_auth="CREDENTIALS")

    assert auth.get_credentials() == "secret-keys"



#!/usr/bin/env python
from pathlib import Path
import sys

from craft_cli import emit, EmitterMode
from craft_store import StoreClient, endpoints
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor


emit.init(EmitterMode.BRIEF, "craft.store-howto", "Starting howto app.")

store_client = StoreClient(
    base_url="https://dashboard.staging.snapcraft.io",
    storage_base_url="https://upload.apps.staging.ubuntu.com",
    endpoints=endpoints.SNAP_STORE,
    user_agent="Craft Store Howto Agent",
    application_name="craft-store-upload",
    environment_auth="SNAPCRAFT_STORE_CREDENTIALS",
)


def create_callback(encoder: MultipartEncoder):
    with emit.progress_bar("Uploading...", encoder.len, delta=False) as progress:

        def progress_callback(monitor: MultipartEncoderMonitor):
            progress.advance(monitor.bytes_read)

    return progress_callback


upload_id = store_client.upload_file(
    monitor_callback=create_callback,
    filepath=Path(sys.argv[1]),
)

emit.message(f"upload-id: {upload_id}")
emit.ended_ok()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023-2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Resource revision models for the Store."""
import collections
import datetime
from enum import Enum
from typing import Annotated, TypeVar

import pydantic

from craft_store.models._base_model import MarshableModel

T = TypeVar("T")


def _validate_list_is_unique(value: list[T]) -> list[T]:
    value_set = set(value)
    if len(value_set) == len(value):
        return value
    dupes = [item for item, count in collections.Counter(value).items() if count > 1]
    raise ValueError(f"Duplicate values in list: {dupes}")


UniqueList = Annotated[
    list[T],
    pydantic.AfterValidator(_validate_list_is_unique),
    pydantic.Field(json_schema_extra={"uniqueItems": True}),
]


class CharmResourceType(str, Enum):
    """Resource types for OCI images."""

    OCI_IMAGE = "oci-image"
    FILE = "file"


class ResponseCharmResourceBase(MarshableModel):
    """A base for a charm resource."""

    name: str = "all"
    channel: str = "all"
    architectures: list[str] = ["all"]


class CharmResourceRevision(MarshableModel):
    """A basic resource revision."""

    bases: list[ResponseCharmResourceBase]
    created_at: datetime.datetime
    name: str
    revision: int
    sha256: str
    sha3_384: str
    sha384: str
    sha512: str
    size: pydantic.ByteSize
    type: CharmResourceType | str
    updated_at: datetime.datetime | None = None
    updated_by: str | None = None


class RequestCharmResourceBase(MarshableModel):
    """A base for a charm resource for use in requests."""

    name: str = "all"
    channel: str = "all"
    architectures: UniqueList[str] = pydantic.Field(
        default_factory=lambda: ["all"], min_length=1
    )


RequestCharmResourceBaseList = Annotated[
    list[RequestCharmResourceBase], pydantic.Field(min_length=1)
]


class CharmResourceRevisionUpdateRequest(MarshableModel):
    """A charm resource revision update request."""

    revision: pydantic.PositiveInt
    bases: RequestCharmResourceBaseList



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Craft Store Authentication Store."""

import base64
import binascii
import json
import logging
import os
import sys
from pathlib import Path
from typing import cast

import keyring
import keyring.backend
import keyring.backends.fail
import keyring.errors
from keyring._compat import properties
from keyring.backends import SecretService
from xdg import BaseDirectory  # type: ignore[import]

from . import errors

logger = logging.getLogger(__name__)


if sys.platform == "linux":
    from secretstorage.exceptions import SecretServiceNotAvailableException

    KEYRING_EXCEPTIONS = (keyring.errors.InitError, SecretServiceNotAvailableException)
else:
    KEYRING_EXCEPTIONS = (keyring.errors.InitError,)


class MemoryKeyring(keyring.backend.KeyringBackend):
    """A keyring that stores credentials in a dictionary."""

    @properties.classproperty  #  type: ignore[misc]
    def priority(self) -> int | float:
        """Supply a priority.

        Indicating the priority of the backend relative to all other backends.
        """
        # Only > 0 make it to the chainer.
        return -1

    def __init__(self) -> None:
        super().__init__()

        self._credentials: dict[tuple[str, str], str] = {}

    def set_password(self, service: str, username: str, password: str) -> None:
        """Set the service password for username in memory."""
        self._credentials[service, username] = password

    def get_password(self, service: str, username: str) -> str | None:
        """Get the service password for username from memory."""
        return self._credentials.get((service, username))

    def delete_password(self, service: str, username: str) -> None:
        """Delete the service password for username from memory."""
        try:
            del self._credentials[service, username]
        except KeyError as key_error:
            raise keyring.errors.PasswordDeleteError from key_error


class FileKeyring(keyring.backend.KeyringBackend):
    """A keyring that stores credentials in a file."""

    @properties.classproperty  #  type: ignore[misc]
    def priority(self) -> int | float:
        """Supply a priority.

        Indicating the priority of the backend relative to all other backends.
        """
        # Only > 0 make it to the chainer.
        return -1

    @property
    def credentials_file(self) -> Path:
        """Credentials file path for instance of application_name."""
        return (
            Path(BaseDirectory.save_data_path(self._application_name))
            / "credentials.json"
        )

    def _write_credentials(self) -> None:
        if not self.credentials_file.exists():
            self.credentials_file.parent.mkdir(parents=True, exist_ok=True)
            self.credentials_file.touch(mode=0o600)
        with self.credentials_file.open("w") as credentials_file:
            json.dump(self._credentials, credentials_file)

    def _read_credentials(self) -> None:
        try:
            with self.credentials_file.open() as credentials_file:
                self._credentials = json.load(credentials_file)
        except (FileNotFoundError, json.decoder.JSONDecodeError):
            self._credentials = {}

    def __init__(self, application_name: str) -> None:
        super().__init__()

        self._application_name = application_name
        self._read_credentials()

    def set_password(self, service: str, username: str, password: str) -> None:
        """Set the service password for username in memory."""
        if service not in self._credentials:
            self._credentials[service] = {username: password}
        else:
            self._credentials[service][username] = password
        self._write_credentials()

    def get_password(self, service: str, username: str) -> str | None:
        """Get the service password for username from memory."""
        try:
            return cast(str, self._credentials[service][username])
        except KeyError:
            return None

    def delete_password(self, service: str, username: str) -> None:
        """Delete the service password for username from memory."""
        try:
            del self._credentials[service][username]
        except KeyError as key_error:
            raise keyring.errors.PasswordDeleteError from key_error
        self._write_credentials()


class Auth:
    """Auth wraps around the keyring to store credentials.

    The application_name and host are used as key/values in the keyring to set,
    get and delete credentials.

    If environment_auth is set on initialization of this class, then a
    :attr:`MemoryKeyring` is setup in lieu of the system one.

    Credentials are base64 encoded into the keyring and decoded on
    retrieval.

    :ivar application_name: name of the application using this library.
    :ivar host: specific host for the store used.
    """

    def __init__(
        self,
        application_name: str,
        host: str,
        ephemeral: bool = False,
        environment_auth: str | None = None,
    ) -> None:
        """Initialize Auth.

        :param application_name: name of the application using this library.
        :param host: specific host for the store used.
        :param environment_auth: environment variable used for authentication.
        :param ephemeral: keep everything in memory.
        """
        self.application_name = application_name
        self.host = host

        environment_auth_value = None
        if environment_auth:
            environment_auth_value = os.getenv(environment_auth)

        if environment_auth_value or ephemeral:
            keyring.set_keyring(MemoryKeyring())

        self._keyring = keyring.get_keyring()
        # This keyring would fail on first use, fail early instead.
        # Only SecretService has get_preferred_collection, which can
        # fail if the keyring backend cannot be unlocked.
        if isinstance(self._keyring, SecretService.Keyring):
            try:
                self._keyring.get_preferred_collection()
            except keyring.errors.KeyringLocked as err:
                raise errors.KeyringUnlockError from err
            except KEYRING_EXCEPTIONS:
                logger.warning("Falling back to file based storage")
                keyring.set_keyring(FileKeyring(application_name))
                self._keyring = keyring.get_keyring()
        elif isinstance(self._keyring, keyring.backends.fail.Keyring):
            raise errors.NoKeyringError

        if environment_auth_value:
            self.set_credentials(self.decode_credentials(environment_auth_value))

    @staticmethod
    def decode_credentials(encoded_credentials: str) -> str:
        """Decode base64 encoded credentials.

        :raises errors.CredentialsNotParseable: when the credentials are incorrectly encoded.
        """
        try:
            return base64.b64decode(encoded_credentials).decode()
        except binascii.Error as err:
            raise errors.CredentialsNotParseable from err

    @staticmethod
    def encode_credentials(credentials: str) -> str:
        """Encode credentials to base64."""
        return base64.b64encode(credentials.encode()).decode()

    def ensure_no_credentials(self) -> None:
        """Check that no credentials exist.

        :raises errors.CredentialsAvailable: if credentials have already been set.
        :raises errors.KeyringUnlockError: if the keyring cannot be unlocked.
        """
        try:
            password = self._keyring.get_password(self.application_name, self.host)
        except keyring.errors.KeyringLocked as exc:
            raise errors.KeyringUnlockError from exc

        if password is not None:
            raise errors.CredentialsAlreadyAvailable(self.application_name, self.host)

    def set_credentials(self, credentials: str, force: bool = False) -> None:
        """Store credentials in the keyring.

        :param credentials: token to store.
        :param force: overwrite existing credentials.
        """
        if not force:
            self.ensure_no_credentials()

        logger.debug(
            "Storing credentials for %r on %r in keyring %r.",
            self.application_name,
            self.host,
            self._keyring.name,
        )
        encoded_credentials = self.encode_credentials(credentials)
        self._keyring.set_password(
            self.application_name, self.host, encoded_credentials
        )

    def get_credentials(self) -> str:
        """Retrieve credentials from the keyring."""
        logger.debug(
            "Retrieving credentials for %r on %r from keyring %r.",
            self.application_name,
            self.host,
            self._keyring.name,
        )

        try:
            encoded_credentials_string = self._keyring.get_password(
                self.application_name, self.host
            )
        except Exception as unknown_error:
            logger.debug(
                "Unhandled exception raised when retrieving credentials: %r",
                unknown_error,
            )
            raise errors.CredentialsUnavailable(
                self.application_name, self.host
            ) from unknown_error

        if encoded_credentials_string is None:
            logger.debug("Credentials not found in the keyring %r", self._keyring.name)
            raise errors.CredentialsUnavailable(self.application_name, self.host)
        return self.decode_credentials(encoded_credentials_string)

    def del_credentials(self) -> None:
        """Delete credentials from the keyring."""
        # Try to get the credentials first to see if there are any,
        # this is to provide an easier troubleshooting experience.
        self.get_credentials()

        logger.debug(
            "Deleting credentials for %r on %r from keyring %r.",
            self.application_name,
            self.host,
            self._keyring.name,
        )
        self._keyring.delete_password(self.application_name, self.host)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

import pytest
from craft_store.models import snap_list_releases_model


@pytest.fixture()
def payload():
    return {
        "channel-map": [
            {
                "channel": "latest/edge",
                "expiration-date": None,
                "progressive": {"paused": None, "percentage": None},
                "resources": [],
                "revision": 5,
                "when": "2021-11-19T11:55:01Z",
                "architecture": "amd64",
            }
        ],
        "package": {
            "channels": [
                {
                    "branch": None,
                    "fallback": None,
                    "name": "latest/stable",
                    "risk": "stable",
                    "track": "latest",
                },
                {
                    "branch": None,
                    "fallback": "latest/stable",
                    "name": "latest/candidate",
                    "risk": "candidate",
                    "track": "latest",
                },
                {
                    "branch": None,
                    "fallback": "latest/candidate",
                    "name": "latest/beta",
                    "risk": "beta",
                    "track": "latest",
                },
                {
                    "branch": None,
                    "fallback": "latest/beta",
                    "name": "latest/edge",
                    "risk": "edge",
                    "track": "latest",
                },
            ]
        },
        "revisions": [
            {
                "architectures": ["amd64"],
                "base": "core22",
                "created-at": "2021-11-19T11:55:00.084876",
                "created-by": "developer",
                "build-url": "https://github.com/actions/build/1",
                "revision": 5,
                "sha3-384": "849fc881f82f076b4a682ddd902a0a13ddbe1ee01200160f38815b3f537db2b943e12faae4ab74f6aa983a26af19d917",
                "size": 18598,
                "status": "released",
                "version": "5",
                "confinement": "strict",
                "grade": "stable",
                "type": "app",
            }
        ],
    }


def test_get_list_releases(payload):
    model = snap_list_releases_model.ListReleasesModel(**payload)

    assert len(model.channel_map) == 1
    assert model.channel_map[0].architecture == "amd64"
    assert model.channel_map[0].channel == "latest/edge"
    assert model.channel_map[0].expiration_date is None
    assert model.channel_map[0].progressive.paused is None
    assert model.channel_map[0].progressive.percentage is None
    assert model.channel_map[0].revision == 5
    assert model.channel_map[0].when == datetime.datetime(
        2021, 11, 19, 11, 55, 1, tzinfo=datetime.timezone.utc
    )

    assert len(model.package.channels) == 4
    assert model.package.channels[0].branch is None
    assert model.package.channels[0].fallback is None
    assert model.package.channels[0].name == "latest/stable"
    assert model.package.channels[0].risk == "stable"
    assert model.package.channels[0].track == "latest"
    assert model.package.channels[1].branch is None
    assert model.package.channels[1].fallback == "latest/stable"
    assert model.package.channels[1].name == "latest/candidate"
    assert model.package.channels[1].risk == "candidate"
    assert model.package.channels[1].track == "latest"
    assert model.package.channels[2].branch is None
    assert model.package.channels[2].fallback == "latest/candidate"
    assert model.package.channels[2].name == "latest/beta"
    assert model.package.channels[2].risk == "beta"
    assert model.package.channels[2].track == "latest"
    assert model.package.channels[3].branch is None
    assert model.package.channels[3].fallback == "latest/beta"
    assert model.package.channels[3].name == "latest/edge"
    assert model.package.channels[3].risk == "edge"
    assert model.package.channels[3].track == "latest"

    assert len(model.revisions) == 1
    assert model.revisions[0].architectures == ["amd64"]
    assert model.revisions[0].base == "core22"
    assert model.revisions[0].created_at == datetime.datetime(
        2021, 11, 19, 11, 55, 0, 84876
    )
    assert model.revisions[0].created_by == "developer"
    assert model.revisions[0].build_url == "https://github.com/actions/build/1"
    assert model.revisions[0].revision == 5
    assert (
        model.revisions[0].sha3_384
        == "849fc881f82f076b4a682ddd902a0a13ddbe1ee01200160f38815b3f537db2b943e12faae4ab74f6aa983a26af19d917"
    )
    assert model.revisions[0].size == 18598
    assert model.revisions[0].status == "released"
    assert model.revisions[0].version == "5"
    assert model.revisions[0].confinement == "strict"
    assert model.revisions[0].grade == "stable"
    assert model.revisions[0].type == "app"



#!/usr/bin/env python3
"""Demo code for uploading a charm to Charmhub."""
import argparse
import pathlib
import sys
import time
from typing import Any

from craft_store import endpoints, StoreClient
from craft_store.models import (
    RevisionsRequestModel,
    ReleaseRequestModel,
    ResourceModel,
    CharmResourceType,
)


def parse_args(argv: list[str]):
    parser = argparse.ArgumentParser(
        prog="upload_package",
        description="Uploads a charm and its resource",
    )
    parser.add_argument(
        "charm_name",
        help="The name of the charm in the store",
    )
    parser.add_argument(
        "resource_name",
        help="The name of the resource in the store",
    )
    parser.add_argument(
        "--charm",
        type=pathlib.Path,
        required=True,
        help="The path of the charm to upload",
    )
    parser.add_argument(
        "--resource",
        type=pathlib.Path,
        required=True,
        help="The path of the resource file to upload",
    )
    return parser.parse_args(argv)


def check_status(client: StoreClient, status_url: str) -> list[dict[str, Any]]:
    """Check the status of an upload."""
    timeout = time.monotonic() + 120
    while time.monotonic() < timeout:
        resource_status = client.request("GET", status_url).json()
        done = True
        for resource_revision_status in resource_status["revisions"]:
            if resource_revision_status["status"] not in ("approved", "rejected"):
                done = False
        if done:
            return resource_status["revisions"]
        time.sleep(3)
    raise TimeoutError("Status was neither approved nor rejected after 120s")


def main(argv: list[str]):
    args = parse_args(argv)
    charm_path = args.charm.expanduser().resolve()
    resource_path = args.resource.expanduser().resolve()
    charm_name = args.charm_name
    resource_name = args.resource_name

    # [docs:get-client]
    base_url: str = "https://api.staging.charmhub.io"
    storage_base_url: str = "https://storage.staging.snapcraftcontent.com"
    client = StoreClient(
        application_name="craft-store-demo",
        base_url=base_url,
        storage_base_url=storage_base_url,
        endpoints=endpoints.CHARMHUB,
        user_agent="craft-store-demo-app",
        environment_auth="CRAFT_STORE_CHARMCRAFT_CREDENTIALS",
    )
    # [docs:get-client-end]
    # [docs:upload-resource]
    resource_upload_id = client.upload_file(filepath=resource_path)
    resource_status_url = client.push_resource(
        name=charm_name,
        resource_name=resource_name,
        upload_id=resource_upload_id,
        resource_type=CharmResourceType.FILE,
    )
    resource_status = check_status(client, base_url + resource_status_url)[0]
    resource_revision = int(resource_status["revision"])
    # [docs:upload-resource-end]
    # [docs:upload-charm]
    charm_upload_id = client.upload_file(filepath=charm_path)
    charm_status_url = client.notify_revision(
        name=charm_name,
        revision_request=RevisionsRequestModel(upload_id=charm_upload_id),
    ).status_url
    charm_status = check_status(client, base_url + charm_status_url)[0]
    charm_revision = charm_status["revision"]
    # [docs:upload-charm-end]

    # [docs:release]
    client.release(
        name=charm_name,
        release_request=[
            ReleaseRequestModel(
                channel="edge",
                resources=[
                    ResourceModel(name=resource_name, revision=resource_revision)
                ],
                revision=charm_revision,
            )
        ],
    )
    # [docs:release-end]


if __name__ == "__main__":
    main(sys.argv[1:])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for update_resource_revisions."""

from craft_store.models.resource_revision_model import (
    CharmResourceRevisionUpdateRequest,
    RequestCharmResourceBase,
)

from .conftest import needs_charmhub_credentials


@needs_charmhub_credentials()
def test_charm_update_resource_revisions(charm_client, charmhub_charm_name):
    resource_name = "empty-file"

    revisions = charm_client.list_resource_revisions(charmhub_charm_name, resource_name)
    revision_numbers = [r.revision for r in revisions[-2:]]
    assert len(revision_numbers) == 2, "Needs more resource revisions"

    resource_updates = [
        CharmResourceRevisionUpdateRequest(
            revision=revision_number,
            bases=[
                RequestCharmResourceBase(
                    name="ubuntu",
                    channel="22.04",
                    architectures=["riscv64"],
                )
            ],
        )
        for revision_number in revision_numbers
    ]
    default_updates = [
        CharmResourceRevisionUpdateRequest(
            revision=revision_number,
            bases=[
                RequestCharmResourceBase(
                    name="all", channel="all", architectures=["all"]
                )
            ],
        )
        for revision_number in revision_numbers
    ]

    update_count = charm_client.update_resource_revisions(
        *resource_updates, name=charmhub_charm_name, resource_name="empty-file"
    )
    assert update_count == len(resource_updates)

    # Reset back to the default.
    update_count = charm_client.update_resource_revisions(
        *default_updates, name=charmhub_charm_name, resource_name="empty-file"
    )
    assert update_count == len(default_updates)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from typing import cast

import pytest
from craft_store.models import revisions_model

from .conftest import needs_charmhub_credentials

pytestmark = pytest.mark.timeout(10)  # Timeout if any test takes over 10 sec.


@needs_charmhub_credentials()
def test_charm_upload(charm_client, fake_charm_file, charmhub_charm_name):
    upload_id = charm_client.upload_file(filepath=fake_charm_file)

    request_model = revisions_model.RevisionsRequestModel(**{"upload-id": upload_id})

    model_response = cast(
        revisions_model.RevisionsResponseModel,
        charm_client.notify_revision(
            name=charmhub_charm_name,
            revision_request=request_model,
        ),
    )

    assert (
        model_response.status_url
        == f"/v1/charm/{charmhub_charm_name}/revisions/review?upload-id={upload_id}"
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from craft_store.errors import StoreServerError

from .conftest import needs_charmhub_credentials

pytestmark = pytest.mark.timeout(10)  # Timeout if any test takes over 10 sec.


@needs_charmhub_credentials()
@pytest.mark.parametrize("entity_type", ["charm", "bundle"])
def test_register_unregister_cycle(charm_client, unregistered_charm_name, entity_type):
    try:
        charm_client.register_name(unregistered_charm_name, entity_type=entity_type)

        names = {result.name for result in charm_client.list_registered_names()}
        assert (
            unregistered_charm_name in names
        ), f"{entity_type} was not successfully registered."
    finally:
        charm_client.unregister_name(unregistered_charm_name)

    names = {result.name for result in charm_client.list_registered_names()}
    assert (
        unregistered_charm_name not in names
    ), f"{entity_type} was not successfully unregistered."


@needs_charmhub_credentials()
def test_unregister_nonexistent(charm_client, unregistered_charm_name):
    with pytest.raises(StoreServerError) as exc_info:
        charm_client.unregister_name(unregistered_charm_name)

    assert "resource-not-found" in exc_info.value.error_list



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import pathlib
import shutil
import uuid
from pathlib import Path

import pytest
import yaml
from craft_store import StoreClient, endpoints


@pytest.fixture()
def charm_client():
    """A common StoreClient for charms"""
    return StoreClient(
        application_name="integration-test",
        base_url="https://api.staging.charmhub.io",
        storage_base_url="https://storage.staging.snapcraftcontent.com",
        endpoints=endpoints.CHARMHUB,
        user_agent="integration-tests",
        environment_auth="CRAFT_STORE_CHARMCRAFT_CREDENTIALS",
    )


@pytest.fixture()
def charmhub_charm_name():
    """Allow overriding the user to override the test charm.

    NOTE: Most integration tests check specifics about craft-store-test-charm,
    so overriding the test charm may cause test failures.
    """
    return os.getenv("CRAFT_STORE_TEST_CHARM", default="craft-store-test")


@pytest.fixture()
def fake_charm_file(tmp_path, charmhub_charm_name):
    """Provide a fake charm to upload to charmhub."""
    return _make_charm(tmp_path, charmhub_charm_name, ["amd64"])


def _make_charm(
    dest_dir: pathlib.Path, name: str, architectures: list[str]
) -> pathlib.Path:
    """Make a fake charm on disk."""
    prime_dir = dest_dir / f"prime_{name}_{'_'.join(architectures)}"
    prime_dir.mkdir()

    metadata_path = prime_dir / "metadata.yaml"
    with metadata_path.open("w") as metadata_file:
        yaml.safe_dump(
            data={
                "name": name,
                "display-name": "display",
                "description": "description",
                "summary": "summary",
                "resources": {
                    "my-rock": {"type": "oci-image"},
                    "my-file": {"type": "file", "filename": "my-file"},
                },
            },
            stream=metadata_file,
        )

    manifest_path = prime_dir / "manifest.yaml"
    with manifest_path.open("w") as manifest_file:
        yaml.safe_dump(
            data={
                "analysis": {
                    "attributes": [
                        {
                            "name": "language",
                            "result": "python",
                        },
                        {
                            "name": "framework",
                            "result": "operator",
                        },
                    ]
                },
                "bases": [
                    {
                        "architectures": architectures,
                        "channel": "22.04",
                        "name": "ubuntu",
                    }
                ],
                "charmcraft-started-at": "2022-04-03T22:27:43.044456Z",
                "charmcraft-version": "1.5.0+12.g04477df.dirty",
            },
            stream=manifest_file,
        )

        charm_file = dest_dir / f"{name}_{'_'.join(architectures)}.charm"
        Path(shutil.make_archive(str(charm_file), "zip", str(prime_dir))).rename(
            charm_file
        )

        shutil.rmtree(prime_dir)

        return charm_file


@pytest.fixture()
def fake_charms(
    tmp_path, charmhub_charm_name, architectures=("amd64", "arm64", "riscv64")
):
    files = [
        _make_charm(tmp_path, charmhub_charm_name, [arch]) for arch in architectures
    ]
    files.append(_make_charm(tmp_path, charmhub_charm_name, architectures))
    return files


@pytest.fixture()
def unregistered_charm_name(charm_client):
    """Get an unregistered name for use in tests"""
    account_id = charm_client.whoami().get("account", {}).get("id", "").lower()
    registered_names = {result.name for result in charm_client.list_registered_names()}
    while (name := f"test-{account_id}-{uuid.uuid4()}") in registered_names:
        # Regenerate UUIDs until we find one that's not registered or timeout.
        pass
    return name


def needs_charmhub_credentials():
    return pytest.mark.skipif(
        not os.getenv("CRAFT_STORE_CHARMCRAFT_CREDENTIALS"),
        reason="CRAFT_STORE_CHARMCRAFT_CREDENTIALS are not set",
    )



#  -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*
#
#  Copyright 2023 Canonical Ltd.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License version 3 as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for methods in the BaseClient class."""
import datetime
from unittest.mock import Mock

import pydantic
import pytest
import requests
from craft_store import BaseClient, endpoints
from craft_store.models import AccountModel, RegisteredNameModel
from craft_store.models._charm_model import CharmBaseModel
from craft_store.models._snap_models import Confinement, Grade, Type
from craft_store.models.resource_revision_model import (
    CharmResourceRevision,
    CharmResourceRevisionUpdateRequest,
    CharmResourceType,
    RequestCharmResourceBase,
    ResponseCharmResourceBase,
)
from craft_store.models.revisions_model import (
    CharmRevisionModel,
    GitRevisionModel,
    SnapRevisionModel,
)

SAMPLE_REAL_REGISTERED_NAMES = b"""{
  "results": [
    {
      "authority": null,
      "contact": null,
      "default-track": null,
      "description": null,
      "id": "test1id",
      "links": {},
      "media": [],
      "name": "test-charm",
      "private": false,
      "publisher": {
        "display-name": "Charmhub Publisher",
        "id": "pubid",
        "username": "usso-someone",
        "validation": "unproven"
      },
      "status": "registered",
      "store": "ubuntu",
      "summary": null,
      "title": null,
      "track-guardrails": [],
      "tracks": [],
      "type": "charm",
      "website": null
    }
  ]
}
"""
DEMO_ACCOUNT = AccountModel(
    id="pubid",
    username="usso-someone",
    validation="unproven",
    **{"display-name": "Charmhub Publisher"},
)


class ConcreteTestClient(BaseClient):
    def _get_authorization_header(self) -> str:
        return "I am authorised."

    def _get_discharged_macaroon(
        self, root_macaroon: str, **kwargs  # noqa: ARG002
    ) -> str:
        return "The voltmeter reads 0V over this macaroon."


@pytest.fixture()
def charm_client():
    client = ConcreteTestClient(
        base_url="https://staging.example.com",
        storage_base_url="https://storage.staging.example.com",
        endpoints=endpoints.CHARMHUB,
        application_name="testcraft",
        user_agent="craft-store unit tests, should not be hitting a real server",
    )
    client.http_client = Mock(spec=client.http_client)
    return client


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (b'{"results":[]}', []),
        (
            SAMPLE_REAL_REGISTERED_NAMES,
            [
                RegisteredNameModel(
                    id="test1id",
                    name="test-charm",
                    private=False,
                    publisher=DEMO_ACCOUNT,
                    status="registered",
                    store="ubuntu",
                    type="charm",
                ),
            ],
        ),
    ],
)
def test_list_registered_names(charm_client, content, expected):
    charm_client.http_client.request.return_value = response = requests.Response()
    response._content = content
    actual = charm_client.list_registered_names()

    assert actual == expected


@pytest.mark.parametrize("resource_type", list(CharmResourceType))
@pytest.mark.parametrize(
    "bases",
    [
        [RequestCharmResourceBase()],
        [
            RequestCharmResourceBase(
                name="ubuntu",
                channel="22.04",
                architectures=["amd64", "arm64", "riscv64"],
            )
        ],
    ],
)
def test_push_resource(
    charm_client,
    resource_type: CharmResourceType,
    bases: list[RequestCharmResourceBase],
):
    name = "my-charm"
    resource_name = "my-resource"
    charm_client.http_client.request.return_value.json.return_value = {
        "status-url": "I am a status URL",
    }

    request_model = {
        "upload-id": "I am an upload",
        "type": resource_type,
        "bases": [b.model_dump(exclude_defaults=False) for b in bases],
    }

    status_url = charm_client.push_resource(
        name,
        resource_name,
        upload_id="I am an upload",
        resource_type=resource_type,
        bases=bases,
    )
    assert status_url == "I am a status URL"

    charm_client.http_client.request.assert_called_once_with(
        "POST",
        "https://staging.example.com/v1/charm/my-charm/resources/my-resource/revisions",
        params=None,
        headers={"Authorization": "I am authorised."},
        json=request_model,
    )


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        pytest.param(b'{"revisions":[]}', [], id="empty"),
        pytest.param(
            b"""{"revisions":[{
                "commit-id": "abc",
                "created-at": "2000-01-01T00:00:00",
                "created-by": "lengau",
                "revision": 1,
                "sha3-384": "its_a_fake",
                "status": "statusy"
            }]}""",
            [
                GitRevisionModel(
                    created_at=datetime.datetime(2000, 1, 1, 0, 0, 0),
                    revision=1,
                    sha3_384="its_a_fake",
                    status="statusy",
                    commit_id="abc",
                    created_by="lengau",
                )
            ],
            id="git-revision",
        ),
        pytest.param(
            b"""{"revisions":[{
                "created-at": "2000-01-01T00:00:00",
                "revision": 1,
                "sha3-384": "its_a_fake",
                "status": "statusy",
                "bases": [{"name": "all", "channel": "all", "architecture": "all"}],
                "size": 1234,
                "version": "1.0.0-0-0"
            }]}""",
            [
                CharmRevisionModel(
                    created_at=datetime.datetime(2000, 1, 1, 0, 0, 0),
                    revision=1,
                    sha3_384="its_a_fake",
                    status="statusy",
                    size=1234,
                    version="1.0.0-0-0",
                    bases=[
                        CharmBaseModel(name="all", channel="all", architecture="all")
                    ],
                )
            ],
            id="charm-revision",
        ),
        pytest.param(
            b"""{"revisions":[{
                "created-at": "2000-01-01T00:00:00",
                "revision": 1,
                "sha3-384": "its_a_fake",
                "status": "statusy",
                "apps": ["appy-mc-app-face"],
                "architectures": ["riscv64"],
                "base": "core24",
                "confinement": "strict",
                "created-by": "lengau",
                "grade": "stable",
                "size": 54321,
                "type": "app",
                "version": "1.2.3"
            }]}""",
            [
                SnapRevisionModel(
                    created_at=datetime.datetime(2000, 1, 1, 0, 0, 0),
                    revision=1,
                    sha3_384="its_a_fake",
                    status="statusy",
                    apps=["appy-mc-app-face"],
                    architectures=["riscv64"],
                    base="core24",
                    confinement=Confinement.STRICT,
                    created_by="lengau",
                    grade=Grade.STABLE,
                    size=54321,
                    type=Type.APP,
                    version="1.2.3",
                )
            ],
            id="snap-revision",
        ),
    ],
)
def test_list_revisions(charm_client, content, expected):
    charm_client.http_client.request.return_value = response = requests.Response()
    response._content = content

    actual = charm_client.list_revisions("my_name")

    assert actual == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        pytest.param(b'{"revisions":[]}', [], id="empty"),
        pytest.param(
            b"""{"revisions":[{
                "bases": [{"name": "all", "channel": "all", "architectures": ["all"]}],
                "created-at": "1970-01-01T00:00:00",
                "name": "resource",
                "revision": 1,
                "sha256": "a-sha256",
                "sha3-384": "a 384-bit sha3",
                "sha384": "a sha384",
                "sha512": "a sha512",
                "size": 17,
                "type": "file",
                "updated-at": "2020-03-14T00:00:00",
                "updated-by": "lengau"
            }]}""",
            [
                CharmResourceRevision(
                    bases=[ResponseCharmResourceBase()],
                    created_at=datetime.datetime(1970, 1, 1),
                    name="resource",
                    revision=1,
                    sha256="a-sha256",
                    sha3_384="a 384-bit sha3",
                    sha384="a sha384",
                    sha512="a sha512",
                    size=pydantic.ByteSize(17),
                    type=CharmResourceType.FILE,
                    updated_at=datetime.datetime(2020, 3, 14),
                    updated_by="lengau",
                )
            ],
        ),
        # Invalid data from the store that we should accept anyway.
        pytest.param(
            b"""{"revisions":[{
                "bases": [
                    {"name": "all", "channel": "all", "architectures": ["all", "all"]},
                    {"name": "all", "channel": "all", "architectures": ["all", "all"]}
                ],
                "created-at": "1970-01-01T00:00:00",
                "name": "",
                "revision": -1,
                "sha256": "",
                "sha3-384": "",
                "sha384": "",
                "sha512": "",
                "size": 0,
                "type": "invalid",
                "updated-at": "2020-03-14T00:00:00",
                "updated-by": ""
            }]}""",
            [
                CharmResourceRevision(
                    bases=[
                        ResponseCharmResourceBase(architectures=["all", "all"]),
                        ResponseCharmResourceBase(architectures=["all", "all"]),
                    ],
                    created_at=datetime.datetime(1970, 1, 1),
                    name="",
                    revision=-1,
                    sha256="",
                    sha3_384="",
                    sha384="",
                    sha512="",
                    size=pydantic.ByteSize(0),
                    type="invalid",
                    updated_at=datetime.datetime(2020, 3, 14),
                    updated_by="",
                ),
            ],
        ),
    ],
)
def test_list_resource_revisions_success(charm_client, content, expected):
    charm_client.http_client.request.return_value = response = requests.Response()
    response._content = content

    actual = charm_client.list_resource_revisions("my-charm", "resource")

    assert actual == expected


def test_list_resource_revisions_not_implemented():
    """list_resource_revisions is not implemented for non-charm namespaces."""
    client = ConcreteTestClient(
        base_url="https://staging.example.com",
        storage_base_url="https://storage.staging.example.com",
        endpoints=endpoints.SNAP_STORE,
        application_name="testcraft",
        user_agent="craft-store unit tests, should not be hitting a real server",
    )
    client.http_client = Mock(spec=client.http_client)

    with pytest.raises(NotImplementedError):
        client.list_resource_revisions("my-snap", "my-resource")


@pytest.mark.parametrize(
    "updates",
    [
        [
            CharmResourceRevisionUpdateRequest(
                revision=1,
                bases=[RequestCharmResourceBase()],
            ),
        ],
    ],
)
def test_update_resource_revisions(charm_client, updates):
    charm_client.http_client.request.return_value.json.return_value = {
        "num-resource-revisions-updated": len(updates)
    }

    actual = charm_client.update_resource_revisions(
        *updates,
        name="my-charm",
        resource_name="my-resource",
    )

    assert actual == len(updates)


def test_update_resource_revisions_empty(charm_client):
    with pytest.raises(
        ValueError, match=r"^Need at least one resource revision to update\.$"
    ):
        charm_client.update_resource_revisions(
            name="my-charm", resource_name="my-resource"
        )


def test_update_resource_revisions_not_implemented():
    """list_resource_revisions is not implemented for non-charm namespaces."""
    client = ConcreteTestClient(
        base_url="https://staging.example.com",
        storage_base_url="https://storage.staging.example.com",
        endpoints=endpoints.SNAP_STORE,
        application_name="testcraft",
        user_agent="craft-store unit tests, should not be hitting a real server",
    )
    client.http_client = Mock(spec=client.http_client)

    with pytest.raises(NotImplementedError):
        client.update_resource_revisions(
            CharmResourceRevisionUpdateRequest(
                revision=1, bases=[RequestCharmResourceBase()]
            ),
            name="my-snap",
            resource_name="my-resource",
        )


def test_update_resource_revision_success(charm_client):
    charm_client.http_client.request.return_value.json.return_value = {
        "num-resource-revisions-updated": 1
    }
    assert charm_client.update_resource_revision(
        "my-charm", "my-resource", revision=1, bases=[RequestCharmResourceBase()]
    )


@pytest.mark.parametrize(
    ("name", "entity_type", "private", "team", "expected_json"),
    [
        pytest.param(
            "test-charm-abcxyz",
            None,
            False,
            None,
            {"name": "test-charm-abcxyz", "private": False},
            id="basic",
        ),
        pytest.param(
            "test-charm-abc123",
            "charm",
            True,
            "starcraft",
            {
                "name": "test-charm-abc123",
                "private": True,
                "type": "charm",
                "team": "starcraft",
            },
            id="all_filled",
        ),
    ],
)
def test_register_name(charm_client, name, entity_type, private, team, expected_json):
    charm_client.request = Mock()
    charm_client.request.return_value.json.return_value = {"id": "abc"}

    charm_client.register_name(
        name, entity_type=entity_type, private=private, team=team
    )

    charm_client.request.assert_called_once_with(
        "POST", "https://staging.example.com/v1/charm", json=expected_json
    )


def test_unregister_name(charm_client):
    charm_client.request = Mock()
    charm_client.request.return_value.json.return_value = {"package-id": "abc"}

    charm_client.unregister_name("test-charm-abcxyz")

    charm_client.request.assert_called_once_with(
        "DELETE", "https://staging.example.com/v1/charm/test-charm-abcxyz"
    )



#  -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*
#
#  Copyright 2023 Canonical Ltd.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License version 3 as published by the Free Software Foundation.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Basic snap-related models."""
from enum import Enum


class Confinement(str, Enum):
    """Snap confinement."""

    CLASSIC = "classic"
    STRICT = "strict"
    DEVMODE = "devmode"


class Grade(str, Enum):
    """Snap grade."""

    DEVEL = "devel"
    STABLE = "stable"


class Type(str, Enum):
    """Type of snap."""

    APP = "app"
    GADGET = "gadget"
    KERNEL = "kernel"
    BASE = "base"
    SNAPD = "snapd"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import cast

from craft_store.models import revisions_model


def test_request_unmarshal_and_marshal():
    payload = {
        "upload-id": "fake-id",
    }

    model = cast(
        revisions_model.RevisionsRequestModel,
        revisions_model.RevisionsRequestModel.unmarshal(payload),
    )

    assert model.upload_id == "fake-id"

    assert model.marshal() == payload


def test_response_unmarshal_and_marshal():
    payload = {
        "status-url": "/foo.bar",
    }

    model = cast(
        revisions_model.RevisionsResponseModel,
        revisions_model.RevisionsResponseModel.unmarshal(payload),
    )

    assert model.status_url == "/foo.bar"

    assert model.marshal() == payload
